import rich.console
import json
import pandas as pd
import sys
import os
import ftplib
import time
import relecov_tools.utils
from datetime import datetime
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule

from ena_upload.ena_upload import extract_targets
from ena_upload.ena_upload import run_construct
from ena_upload.ena_upload import construct_submission
from ena_upload.ena_upload import send_schemas
from ena_upload.ena_upload import process_receipt
from ena_upload.ena_upload import update_table
from ena_upload.ena_upload import update_table_simple

pd.options.mode.chained_assignment = None

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class EnaUpload(BaseModule):
    def __init__(
        self,
        user=None,
        password=None,
        center=None,
        ena_json=None,
        template_path=None,
        dev=None,
        action=None,
        metadata_types=None,
        upload_fastq=None,
        output_dir=None,
    ):
        super().__init__(output_dir=output_dir, called_module="upload-to-ena")
        if user is None:
            self.user = relecov_tools.utils.prompt_text(
                msg="Enter your username defined in ENA"
            )
        else:
            self.user = user
        if password is None:
            self.passwd = relecov_tools.utils.prompt_password(
                msg="Enter your password to ENA"
            )
        else:
            self.passwd = password
        if center is None:
            self.center = relecov_tools.utils.prompt_text(msg="Enter your center name")
        else:
            self.center = center
        if ena_json is None:
            self.source_json_file = relecov_tools.utils.prompt_path(
                msg="Select the ENA json file to upload"
            )
        else:
            self.source_json_file = ena_json
        if not os.path.exists(self.source_json_file):
            self.log.error("json data file %s does not exist ", self.source_json_file)
            stderr.print(f"[red]json data file {self.source_json_file} does not exist")
            sys.exit(1)
        if template_path is None:
            self.template_path = relecov_tools.utils.prompt_path(
                msg="Select the folder containing ENA templates"
            )
        # e.g. template_folder = "/home/user/github_repositories/relecov-tools/relecov_tools/assets/ena_templates"
        else:
            self.template_path = template_path
        if not os.path.exists(self.template_path):
            stderr.print("[red]Error: ENA template folder does not exist")
            sys.exit(1)
        if dev is None:
            self.dev = relecov_tools.utils.prompt_yn_question(
                msg="Do you want to test upload data?"
            )
        else:
            self.dev = dev
        if action is None:
            self.action = relecov_tools.utils.prompt_selection(
                msg="Select the action to upload to ENA",
                choices=["ADD", "MODIFY", "CANCEL", "RELEASE"],
            )
        elif action.upper() not in ["ADD", "MODIFY", "CANCEL", "RELEASE"]:
            stderr.print(f"[red] Action '{action}' not supported")
            sys.exit(1)
        else:
            self.action = action.upper()

        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the folder to store the xml files"
            )
        else:
            self.output_dir = output_dir

        self.upload_fastq_files = upload_fastq

        all_metadata_types = ["study", "run", "experiment", "sample"]
        if metadata_types is None:
            # If not specified, all metadata xmls are generated and submitted
            self.metadata_types = all_metadata_types
        else:
            self.metadata_types = metadata_types.split(",")
            if not all(xml in all_metadata_types for xml in self.metadata_types):
                wrong_types = [
                    xml for xml in self.metadata_types if xml not in all_metadata_types
                ]
                self.log.error("Unsupported metadata xml types: " + str(wrong_types))
                stderr.print(f"[red]Unsupported metadata xml types: {wrong_types}")
                sys.exit(1)

        config_json = ConfigJson()
        self.config_json = config_json
        self.checklist = self.config_json.get_configuration("ENA_fields")["checklist"]
        with open(self.source_json_file, "r") as fh:
            json_data = json.loads(fh.read())
            self.json_data = json_data
        batch_id = self.get_batch_id_from_data(self.json_data)
        self.set_batch_id(batch_id)
        if self.dev:
            self.url = "https://wwwdev.ebi.ac.uk/ena/submit/drop-box/submit/?auth=ENA"
        else:
            self.url = "https://www.ebi.ac.uk/ena/submit/drop-box/submit/?auth=ENA"

    def table_formatting(self, schemas_dataframe_raw, source):
        """Some fields in the dataframe need special formatting"""
        formated_df = schemas_dataframe_raw[source]
        formated_df.insert(3, "status", self.action)
        formated_df.rename(
            columns={
                (str(source) + "_alias"): "alias",
                (str(source) + "_title"): "title",
            },
            inplace=True,
        )
        if self.action in ["CANCEL", "MODIFY", "RELEASE"]:
            formated_df.rename(
                columns={"ena_" + str(source) + "_accession": "accession"}, inplace=True
            )
        if source == "study":
            formated_df = formated_df.drop_duplicates(subset=["alias"])
        if source == "sample":
            formated_df.insert(4, "ENA_CHECKLIST", self.checklist)
        """
        file_name and file_checksum are fields with a structure like this:
        file_nameR1--file_nameR2 / file_checksumR1--file_checksumR2
        The run table needs a row for each strand, so these fields are splitted
        """
        if source == "run":
            formated_df["file_name"] = formated_df["file_name"].str.split("--")
            formated_df = formated_df.explode("file_name").reset_index(drop=True)
            formated_df["file_checksum"] = [
                x.split("--")[0] if index % 2 == 0 else x.split("--")[1]
                for index, x in enumerate(formated_df["file_checksum"])
            ]

        return formated_df

    def dataframes_from_json(self, json_data):
        """The xml is built using a dictionary of dataframes as a base structure"""
        source_options = self.metadata_types
        schemas_dataframe = {}
        schemas_dataframe_raw = {}
        acces_fields = self.config_json.get_topic_data("ENA_fields", "accession_fields")
        filtered_access_fields = [
            fd for fd in acces_fields if any(source in fd for source in source_options)
        ]
        all_missing_accessions = []
        if self.action in ["CANCEL", "MODIFY", "RELEASE"]:
            for source in source_options:
                missing_accessions = [
                    samp["sample_title"]
                    for samp in json_data
                    for fd in filtered_access_fields
                    if (source in fd and fd not in samp.keys())
                ]
                if missing_accessions:
                    self.log.error(
                        "Found samples in json without proper ena accessions"
                    )
                    stderr.print(f"[red]Found samples missing {source} accession ids:")
                all_missing_accessions.extend(missing_accessions)
            if all_missing_accessions:
                stderr.print("Not committed samples:\n", all_missing_accessions)

        for source in source_options:
            source_topic = "_".join(["df", source, "fields"])
            source_fields = self.config_json.get_topic_data("ENA_fields", source_topic)
            if self.action in ["CANCEL", "MODIFY", "RELEASE"]:
                source_fields.append(str("ena_" + source + "_accession"))
            source_dict = {
                field: [
                    sample[field]
                    for sample in json_data
                    if sample["sample_title"] not in all_missing_accessions
                ]
                for field in source_fields
            }
            schemas_dataframe_raw[source] = pd.DataFrame.from_dict(source_dict)
            schemas_dataframe[source] = self.table_formatting(
                schemas_dataframe_raw, source
            )
        return schemas_dataframe

    def save_tables(self, schemas_dataframe, date):
        """Save the dataframes into csv files"""
        stderr.print(f"Saving dataframes in {self.output_dir}")
        for source, table in schemas_dataframe.items():
            table_name = str(self.output_dir + source + date + "_table.csv")
            table.to_csv(table_name, sep=",")

    def update_json(self, updated_schemas_df, json_data):
        access_dict = {}
        updated_json_data = json_data.copy()
        for source, table in updated_schemas_df.items():
            access_list = [x for x in table["accession"]]
            access_dict[source] = access_list
            """run accessions are duplicated for R1/R2 so they need to be removed"""
            if source == "run":
                del access_dict[source][1::2]
        for source, acclist in access_dict.items():
            accession_field_name = str("ena_" + source + "_accession")
            for sample, accession in zip(updated_json_data, acclist):
                sample[accession_field_name] = accession
        return updated_json_data

    def xml_submission(self, json_data, schemas_dataframe, batch_index=None):
        """The metadata is submitted in an xml format"""
        schema_targets = extract_targets(self.action, schemas_dataframe)

        tool = self.config_json.get_configuration("ENA_fields")["tool"]

        if self.action in ["ADD", "MODIFY"]:
            schema_xmls = run_construct(
                self.template_path, schema_targets, self.center, self.checklist, tool
            )
            submission_xml = construct_submission(
                self.template_path,
                self.action,
                schema_xmls,
                self.center,
                self.checklist,
                tool,
            )
        elif self.action in ["CANCEL", "RELEASE"]:
            # when CANCEL/RELEASE, only the accessions are needed
            # schema_xmls is only used to record the following 'submission'
            schema_xmls = {}
            submission_xml = construct_submission(
                self.template_path,
                self.action,
                schema_targets,
                self.center,
                self.checklist,
                tool,
            )
        schema_xmls["submission"] = submission_xml

        stderr.print(f"\nProcessing submission to ENA server: {self.url}")

        receipt = send_schemas(schema_xmls, self.url, self.user, self.passwd).text
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)
        date = str(datetime.now().strftime("%Y%m%d-%H%M%S"))
        receipt_name = "receipt_" + date + ".xml"
        receipt_dir = os.path.join(self.output_dir, receipt_name)
        stderr.print(f"Printing receipt to {receipt_dir}")

        with open(f"{receipt_dir}", "w") as fw:
            fw.write(receipt)
        try:
            schema_update = process_receipt(receipt.encode("utf-8"), self.action)
        except ValueError:
            self.log.error("There was an ERROR during submission:")
            sys.exit(receipt)
        if str(self.action) in ["ADD", "MODIFY"]:
            updated_schemas_df = update_table(
                schemas_dataframe, schema_targets, schema_update
            )
        else:
            updated_schemas_df = update_table_simple(
                schemas_dataframe, schema_targets, self.action
            )

        updated_json = self.update_json(updated_schemas_df, json_data)
        if batch_index is None:
            suffix = str("_" + date + ".json")
        else:
            suffix = str("_" + date + "_batch" + str(batch_index) + ".json")
        updated_json_name = (
            os.path.splitext(os.path.basename(self.source_json_file))[0] + suffix
        )
        relecov_tools.utils.write_json_to_file(updated_json, updated_json_name)

        self.save_tables(updated_schemas_df, date)
        return

    def fastq_submission(self, json_data, max_retries=3, retry_delay=5):
        """The fastq files are submitted apart from the metadata"""
        stderr.print("Submitting fastq files")
        json_dataframe = pd.DataFrame(json_data)
        file_paths = {}
        file_paths_r2 = {}
        for path in json_dataframe["r1_fastq_filepath"]:
            file_paths[os.path.basename(path)] = os.path.abspath(path)

        for path in json_dataframe["r2_fastq_filepath"]:
            file_paths_r2[os.path.basename(path)] = os.path.abspath(path)

        file_paths.update(file_paths_r2)

        connection_retries = 0
        while connection_retries < 3:
            try:
                session = ftplib.FTP(
                    "webin2.ebi.ac.uk", self.user, self.passwd, timeout=60
                )
                session.login(self.user, self.passwd)
                break
            except ftplib.all_errors as e:
                stderr.print(
                    f"Connection attempt {connection_retries+1} failed: {e}. Retrying..."
                )
                connection_retries += 1
                time.sleep(retry_delay)
        self.upload_files_with_retries(session, file_paths, max_retries, retry_delay)

        try:
            session.quit()
        except ftplib.all_errors as e:
            stderr.print(f"ERROR: Could not close FTP session properly: {e}")

    def upload_files_with_retries(
        self, session, file_paths, max_retries=3, retry_delay=5
    ):
        for filename, path in file_paths.items():
            stderr.print(f"Uploading path: {path} with filename: {filename}")
            retries = 0
            while retries < max_retries:
                try:
                    # Reopen file in each retry
                    with open(path, "rb") as file:
                        response = session.storbinary(f"STOR {filename}", file)

                        # Verify if the upload was successful (using response code)
                        if response.startswith("226"):
                            stderr.print(f"File {filename} uploaded successfully.")
                            break
                        else:
                            raise ftplib.error_perm(f"Unexpected response: {response}")
                except (
                    ftplib.error_temp,
                    ftplib.error_perm,
                    ftplib.socket.error,
                    ftplib.error_proto,
                ) as e:
                    # Error handling related to FTP
                    retries += 1
                    stderr.print(f"FTP error: {e}. Retry {retries}/{max_retries}")
                    # Reopen connection if failed due to timeout or temporary error
                    if (
                        isinstance(e, ftplib.error_temp)
                        or "timed out" in str(e).lower()
                    ):
                        stderr.print("Reinitializing FTP connection...")
                        try:
                            session.quit()
                        except Exception:
                            pass
                        time.sleep(retry_delay)

                        connection_retries = 0
                        while connection_retries < 3:
                            try:
                                session = ftplib.FTP(
                                    "webin2.ebi.ac.uk",
                                    self.user,
                                    self.passwd,
                                    timeout=60,
                                )
                                session.login(self.user, self.passwd)
                                break
                            except ftplib.all_errors as e:
                                stderr.print(
                                    f"Connection attempt {connection_retries+1} failed: {e}. Retrying..."
                                )
                                connection_retries += 1
                                time.sleep(retry_delay)
                        if connection_retries == 3:
                            stderr.print(f"Failed to reconnect after {3} attempts.")
                            break
                except Exception as e:
                    # Handling of any other unexpected errors
                    retries += 1
                    stderr.print(
                        f"Unexpected error: {e}. Retry {retries}/{max_retries}"
                    )
                    time.sleep(retry_delay)
            else:
                stderr.print(
                    f"Failed to upload {filename} after {max_retries} retries."
                )
                session.quit()

    def large_json_upload(self, json_data):
        """
        Split large json into smaller jsons of maximum size 20
        due to limitations in submissions to ENA's API
        """
        ena_api_limit = 20
        number_of_batchs = len(range(0, len(json_data), ena_api_limit))
        stderr.print(f"Splitting the json data in {number_of_batchs} batchs...")
        for index in range(0, len(json_data), ena_api_limit):
            batch_index = json_data[index : index + ena_api_limit]
            stderr.print(f"[blue]Processing batch {batch_index}...")
            self.standard_upload(batch_index)
        return

    def standard_upload(self, json_data, batch_index=None):
        """Create the required files and upload to ENA"""
        schemas_dataframe = self.dataframes_from_json(json_data)
        stderr.print("[blue]Successfull creation of dataframes")
        if self.upload_fastq_files:
            self.fastq_submission(json_data)
        stderr.print("Preparing xml files for submission...")
        self.xml_submission(json_data, schemas_dataframe, batch_index)
        return

    def upload(self):
        """Handle the data and upload it to ENA"""
        if len(self.json_data) <= 50:
            self.standard_upload(self.json_data)
        else:
            stderr.print("[yellow]Json is too large to be submitted. Splitting it...")
            self.large_json_upload(self.json_data)
        stderr.print("[green] Finished execution")
        return



#!/usr/bin/env python
import os
import sys
import argparse
from relecov_tools.download_manager import DownloadManager


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--download_option",
        type=str,
        help="Download option",
    )
    parser.add_argument("-t", "--target_folders", type=str, help="Target folders")
    args = parser.parse_args()

    val_dict = {
        "user": os.environ["TEST_USER"],
        "password": os.environ["TEST_PASSWORD"],
        "download_option": args.download_option,
        "output_dir": os.environ["OUTPUT_LOCATION"],
        "target_folders": args.target_folders,
    }
    prepare_remote_test(**val_dict)


def prepare_remote_test(**kwargs):
    print("Initiating sftp module")
    download_manager = DownloadManager(
        user=kwargs["user"],
        password=kwargs["password"],
        conf_file=None,
        download_option=kwargs["download_option"],
        output_dir=kwargs["output_dir"],
        target_folders=kwargs["target_folders"],
        subfolder="RELECOV",
    )

    print("Opening connection to sftp")
    download_manager.relecov_sftp.sftp_port = os.environ["TEST_PORT"]
    if not download_manager.relecov_sftp.open_connection():
        print("Could not open connection to remote sftp")
        sys.exit(1)

    print("Cleaning folders inside RELECOV")
    base_paths = ["COD-test-1/RELECOV", "COD-test-2/RELECOV"]
    for base_path in base_paths:
        remote_folders = download_manager.relecov_sftp.list_remote_folders(
            base_path, recursive=True
        )
        remote_folders = sorted(remote_folders, reverse=True)
        for folder in remote_folders:
            folder = folder.replace("./", "")
            if folder == base_path:
                continue

            filelist = download_manager.relecov_sftp.get_file_list(folder)
            for file in filelist:
                print(f"Removing file {file}")
                download_manager.relecov_sftp.remove_file(file)

            if "tmp_processing" in folder or "invalid_samples" in folder:
                try:
                    print(f"Removing special folder {folder}")
                    download_manager.relecov_sftp.remove_dir(folder)
                except Exception as e:
                    print(f"Could not remove folder {folder}: {e}")

        filelist_base = download_manager.relecov_sftp.get_file_list(base_path)
        for file in filelist_base:
            print(f"Removing file {file} in {base_path}")
            download_manager.relecov_sftp.remove_file(file)

    data_loc = "tests/data/sftp_handle"
    folder_files_dict = {folder: files for folder, _, files in os.walk(data_loc)}
    print("Uploading files to sftp...")
    for folder, files in folder_files_dict.items():
        if "datatest" in folder:
            remote_dir = "COD-test-1/RELECOV"
        elif "empty_test" in folder:
            remote_dir = "COD-test-2/RELECOV"
        else:
            continue
        print(f"Uploading files from {folder}")
        for file in files:
            remote_path = os.path.join(remote_dir, file)
            local_path = os.path.join(os.path.abspath(folder), file)
            download_manager.relecov_sftp.upload_file(local_path, remote_path)

    download_manager.relecov_sftp.close_connection()

    def test_download(download_manager):
        download_manager.execute_process()

    test_download(download_manager)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import os
import sys
import rich.console
import re
import shutil
import numpy as np
from datetime import datetime
from rich.prompt import Prompt

import pandas as pd
import relecov_tools.utils
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class BioinfoReportLog:
    def __init__(self, log_report=None, output_dir="/tmp/"):
        if not log_report:
            self.report = {"error": {}, "valid": {}, "warning": {}}
        else:
            self.report = log_report

    def update_log_report(self, method_name, status, message):
        """Update the progress log report with the given method name, status, and message.

        Args:
            method_name (str): The name of the method being logged.
            status (str): The status of the log message, can be one of 'valid', 'error', or 'warning'.
            message (str): The message to be logged.

        Returns:
            dict: The updated progress log report.

        Raises:
            ValueError: If an invalid status is provided.
        """
        if status == "valid":
            self.report["valid"].setdefault(method_name, []).append(message)
            return self.report
        elif status == "error":
            self.report["error"].setdefault(method_name, []).append(message)
            return self.report
        elif status == "warning":
            self.report["warning"].setdefault(method_name, []).append(message)
            return self.report
        else:
            raise ValueError("Invalid status provided.")

    def print_log_report(self, name, sections):
        """Prints the log report by calling util's function.

        Args:
            name (str): The name of the log report.
            sections (list of str): The sections of the log report to be printed.

        Returns:
            None
        """
        relecov_tools.utils.print_log_report(self.report, name, sections)


# TODO: Add method to validate bioinfo_config.json file requirements.
class BioinfoMetadata(BaseModule):
    def __init__(
        self,
        json_file=None,
        json_schema_file=None,
        input_folder=None,
        output_dir=None,
        software_name=None,
        update=False,
        **kwargs,
    ):
        super().__init__(output_dir=output_dir, called_module=__name__)
        config_json = ConfigJson()
        self.log.info("Initiating read-bioinfo-metadata process")
        # Init process log

        if json_schema_file is None:
            schema_name = config_json.get_topic_data("json_schemas", "relecov_schema")
            json_schema_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "schema", schema_name
            )

        self.json_schema = relecov_tools.utils.read_json_file(json_schema_file)

        if output_dir is None:
            self.output_dir = relecov_tools.utils.prompt_path(
                msg="Select the output folder"
            )
        else:
            self.output_dir = os.path.realpath(output_dir)
        self.logsum = self.parent_log_summary(output_dir=output_dir)
        self.log_report = BioinfoReportLog(output_dir=output_dir)

        # Parse read-lab-meta-data
        if json_file is None:
            json_file = relecov_tools.utils.prompt_path(
                msg="Select the json file that was created by the read-lab-metadata"
            )
        if not os.path.isfile(json_file):
            self.update_all_logs(
                self.__init__.__name__,
                "error",
                f"file {json_file} does not exist",
            )
            sys.exit(
                self.log_report.print_log_report(self.__init__.__name__, ["error"])
            )

        #  Assign a new name for better readability
        self.readlabmeta_json_file = json_file

        # Initialize j_data object
        stderr.print("[blue]Reading lab metadata json")
        self.j_data = self.collect_info_from_lab_json()
        batch_id = self.get_batch_id_from_data(self.j_data)
        self.set_batch_id(batch_id)

        meta_basename = os.path.basename(self.readlabmeta_json_file).split(".")[0]
        if batch_id in meta_basename:
            # File is probably tagged already from a previous process
            self.out_filename = "bioinfo_" + meta_basename + ".json"
        else:
            self.out_filename = self.tag_filename("bioinfo_" + meta_basename + ".json")
        self.update = update

        # Parse input/output folder
        if input_folder is None:
            self.input_folder = relecov_tools.utils.prompt_path(
                msg="Select the input folder"
            )
        else:
            self.input_folder = input_folder

        # Parse bioinfo configuration
        self.bioinfo_json_file = os.path.join(
            os.path.dirname(__file__), "conf", "bioinfo_config.json"
        )
        if software_name is None:
            software_name = relecov_tools.utils.prompt_path(
                msg="Select the software, pipeline or tool use in the bioinformatic analysis: "
            )
        self.software_name = software_name
        available_software = relecov_tools.utils.get_available_software(
            self.bioinfo_json_file
        )
        bioinfo_config = ConfigJson(self.bioinfo_json_file)

        self.schema_path = os.path.join(
            os.path.dirname(__file__), "schema", "relecov_schema.json"
        )
        self.bioinfo_schema = relecov_tools.utils.load_schema(self.schema_path)

        if self.software_name in available_software:
            self.software_config = bioinfo_config.get_configuration(self.software_name)
        else:
            self.update_all_logs(
                self.__init__.__name__,
                "error",
                f"No configuration available for '{self.software_name}'. Currently, the only available software options are:: {', '.join(available_software)}",
            )
            sys.exit(
                self.log_report.print_log_report(self.__init__.__name__, ["error"])
            )

    def update_all_logs(self, method_name, status, message):
        report = self.log_report.update_log_report(method_name, status, message)
        if status == "error":
            self.logsum.add_error(key=method_name, entry=message)
        elif status == "warning":
            self.logsum.add_warning(key=method_name, entry=message)
        return report

    def scann_directory(self):
        """Scanns bioinfo analysis directory and identifies files according to the file name patterns defined in the software configuration json.

        Returns:
            files_found: A dictionary containing file paths found based on the definitions provided in the bioinformatic JSON file within the software scope (self.software_config).
        """
        method_name = f"{self.scann_directory.__name__}"
        total_files = 0
        files_found = {}
        all_scanned_things = []
        for root, dirs, files in os.walk(self.input_folder, topdown=True):
            # Skipping nextflow's work junk directory
            dirs[:] = [d for d in dirs if "work" not in root.split("/")]
            total_files = total_files + len(files)
            all_scanned_things.append((root, files))
        for topic_key, topic_scope in self.software_config.items():
            if "fn" not in topic_scope:  # try/except fn
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"No 'fn' (file pattern) found in '{self.software_name}.{topic_key}'.",
                )
                continue
            for tup in all_scanned_things:
                matching_files = [
                    os.path.join(tup[0], file_name)
                    for file_name in tup[1]
                    if re.search(topic_scope["fn"], os.path.join(tup[0], file_name))
                ]
                if matching_files:
                    if topic_key not in files_found:
                        files_found[topic_key] = []
                    files_found[topic_key].extend(matching_files)
        if len(files_found) < 1:
            self.update_all_logs(
                method_name,
                "error",
                f"No files found in '{self.input_folder}' according to '{os.path.basename(self.bioinfo_json_file)}' file name patterns.",
            )
            sys.exit(self.log_report.print_log_report(method_name, ["error"]))
        else:
            self.update_all_logs(
                self.scann_directory.__name__,
                "valid",
                f"Scanning process succeed. Scanned {total_files} files.",
            )
            self.log_report.print_log_report(method_name, ["valid", "warning"])
            return files_found

    def mapping_over_table(self, j_data, map_data, mapping_fields, table_name):
        """Maps bioinformatics metadata from map_data to j_data based on the mapping_fields.
        Args:
            j_data (list(dict{str:str}): A list of dictionaries containing metadata lab (one item per sample).
            map_data (dict(dict{str:str})): A dictionary containing bioinfo metadata handled by the method handling_files().
            mapping_fields (dict{str:str}): A dictionary of mapping fields defined in the 'content' definition under each software scope (see conf/bioinfo.config).
            table_name (str): Path to the mapping file/table.
        Returns:
            j_data: updated j_data with bioinformatic metadata mapped in it.
        """
        method_name = f"{self.mapping_over_table.__name__}:{self.software_name}.{self.current_config_key}"
        errors = []
        field_errors = {}
        field_valid = {}

        sample_ids = {
            row.get("sequencing_sample_id")
            for row in j_data
            if row.get("sequencing_sample_id")
        }
        matched_samples = sample_ids & set(map_data.keys())

        if not matched_samples:
            no_samples = True
        else:
            no_samples = False

        for row in j_data:
            # TODO: We should consider an independent module that verifies that sample's name matches this pattern.
            #       If we add warnings within this module, every time mapping_over_table is invoked it will print redundant warings
            if not row.get("sequencing_sample_id"):
                self.update_all_logs(
                    method_name,
                    "warning",
                    f'Sequencing_sample_id missing in {row.get("collecting_sample_id")}... Skipping...',
                )
                continue
            sample_name = row["sequencing_sample_id"]
            if sample_name in map_data.keys():
                for field, value in mapping_fields.items():
                    try:
                        raw_val = map_data[sample_name][value]
                        raw_val = self.replace_na_value_if_needed(field, raw_val)
                        expected_type = (
                            self.bioinfo_schema["properties"]
                            .get(field, {})
                            .get("type", "string")
                        )
                        row[field] = relecov_tools.utils.cast_value_to_schema_type(
                            raw_val, expected_type
                        )
                        field_valid[sample_name] = {field: value}
                    except KeyError as e:
                        field_errors[sample_name] = {field: e}
                        row[field] = "Not Provided [SNOMED:434941000124101]"
                        continue
            elif (
                not self.software_config[self.current_config_key].get(
                    "multiple_samples"
                )
                and no_samples is True
            ):  # When sample ID is not in mapping_fields because is not a multiple_sample table
                for field, value_dict in mapping_fields.items():
                    for json_field, software_key in value_dict.items():
                        try:
                            raw_val = map_data[software_key][field]
                            expected_type = (
                                self.bioinfo_schema["properties"]
                                .get(json_field, {})
                                .get("type", "string")
                            )

                            row[json_field] = (
                                relecov_tools.utils.cast_value_to_schema_type(
                                    raw_val, expected_type
                                )
                            )
                            field_valid[software_key] = {json_field: field}
                        except KeyError as e:
                            field_errors[software_key] = {json_field: str(e)}
                            row[json_field] = "Not Provided [SNOMED:434941000124101]"
            else:
                errors.append(sample_name)
                for field in mapping_fields.keys():
                    row[field] = "Not Provided [SNOMED:434941000124101]"
        # work around when map_data comes from several per-sample tables/files instead of single table
        if len(table_name) > 2:
            table_name = os.path.dirname(table_name[0])
        else:
            table_name = table_name[0]
        # Parse missing sample errors
        if errors:
            lenerrs = len(errors)
            self.update_all_logs(
                method_name,
                "warning",
                f"{lenerrs} samples missing in '{table_name}': {', '.join(errors)}.",
            )
        else:
            self.update_all_logs(
                method_name,
                "valid",
                f"All samples were successfully found in {table_name}.",
            )
        # Parse missing fields errors
        # TODO: this stdout can be improved
        if len(field_errors) > 0:
            self.update_all_logs(
                method_name,
                "warning",
                f"Missing fields in {table_name}:\n\t{field_errors}",
            )
        else:
            self.update_all_logs(
                method_name,
                "valid",
                f"Successfully mapped fields in {', '.join(field_valid.keys())}.",
            )
        # Print report
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data

    def validate_samplenames(self):
        """Validate that the sequencing_sample_id from the JSON input is present in the samples_id.txt.

        Raises:
            ValueError: If no sample from the JSON input matches the samples in the samples_id.txt.
        """
        samplesid_path = os.path.join(self.input_folder, "samples_id.txt")
        with open(samplesid_path, "r") as file:
            samplesid_list = [line.strip() for line in file.readlines()]
        json_samples = [sample["sequencing_sample_id"] for sample in self.j_data]
        matching_samples = set(json_samples).intersection(samplesid_list)
        if not matching_samples:
            raise ValueError(
                "No sample from the JSON input matches the samples in the provided analysis folder."
            )
        else:
            print(
                f"Found {len(matching_samples)}/{len(json_samples)} matching samples in the samplesheet."
            )
            self.log.info(
                f"Found {len(matching_samples)}/{len(json_samples)} matching samples in the samplesheet."
            )

    def validate_software_mandatory_files(self, files_dict):
        """Validates the presence of all mandatory files as defined in the software configuration JSON.

        Args:
            files_dict (dict{str:str}): A dictionary containing file paths found based on the definitions provided in the bioinformatic JSON file within the software scope (self.software_config).
        """
        method_name = f"{self.validate_software_mandatory_files.__name__}"
        missing_required = []
        for key in self.software_config:
            if key == "fixed_values":
                continue
            if self.software_config[key].get("required") is True:
                try:
                    files_dict[key]
                except KeyError:
                    missing_required.append(key)
                    continue
            else:
                continue
        if len(missing_required) >= 1:
            self.update_all_logs(
                method_name,
                "error",
                f"Missing mandatory files in {self.software_name}:{', '.join(missing_required)}",
            )
            sys.exit(self.log_report.print_log_report(method_name, ["error"]))
        else:
            self.update_all_logs(
                method_name, "valid", "Successfull validation of mandatory files."
            )
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return

    def add_bioinfo_results_metadata(
        self, files_dict, j_data, sufix, file_tag, output_dir=None
    ):
        """Adds metadata from bioinformatics results to j_data.
        It first calls file_handlers and then maps the handled
        data into j_data.

        Args:
            files_dict (dict{str:str}): A dictionary containing file paths found based on the definitions provided in the bioinformatic JSON file within the software scope (self.software_config).
            j_data (list(dict{str:str}): A list of dictionaries containing metadata lab (list item per sample).
            sufix (str): Sufix added to splitted tables file name.
            output_dir (str): Path to save output files generated during handling_files() process.
            file_tag(str): Tag that will be used for output filenames includes batch date (same as download date) and hex.

        Returns:
            j_data_mapped: A list of dictionaries with bioinformatics metadata mapped into j_data.
        """
        method_name = f"{self.add_bioinfo_results_metadata.__name__}"
        for key in self.software_config.keys():
            # Update bioinfo cofiguration key/scope
            self.current_config_key = key
            map_method_name = f"{method_name}:{self.software_name}.{key}"
            # This skip files that will be parsed with other methods
            if key == "workflow_summary" or key == "fixed_values":
                continue
            try:
                files_dict[key]
                stderr.print(f"[blue]Start processing {self.software_name}.{key}")
                self.log.info(f"Start processing {self.software_name}.{key}")
            except KeyError:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"No file path found for '{self.software_name}.{key}'",
                )
                continue
            # Handling files
            if self.software_config[key].get("map", True) is False:
                msg = f"File '{self.software_name}.{key}' was processed but skipped from mapping as defined in config."
                self.update_all_logs(map_method_name, "warning", msg)
                self.log_report.print_log_report(map_method_name, ["warning"])
                continue
            data_to_map = self.handling_files(
                files_dict[key], sufix, output_dir, file_tag
            )
            # Mapping data to j_data
            mapping_fields = self.software_config[key].get("content")
            if not mapping_fields:
                self.update_all_logs(
                    map_method_name,
                    "warning",
                    f"No metadata found to perform mapping from '{self.software_name}.{key}' despite 'content' fields being defined.",
                )
                self.log_report.print_log_report(map_method_name, ["warning"])
                continue
            if data_to_map:
                j_data_mapped = self.mapping_over_table(
                    j_data=j_data,
                    map_data=data_to_map,
                    mapping_fields=self.software_config[key]["content"],
                    table_name=files_dict[key],
                )
            else:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"No metadata found to perform standard mapping when processing '{self.software_name}.{key}'",
                )
                continue
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data_mapped

    def handling_tables(self, file_list, conf_tab_name):
        """Reads a tabular file in different formats and returns a dictionary containing
        the corresponding data for each sample.

        Args:
            file_list (list): A list of file path/s to be processed.
            conf_tab_name (str): Name of the table in the defined config file.

        Returns:
            data (dict): A dictionary containing metadata as defined in handling_files.
        """
        method_name = f"{self.add_bioinfo_results_metadata.__name__}:{self.handling_tables.__name__}"
        file_ext = os.path.splitext(conf_tab_name)[1]
        sample_idx_colpos = self.get_sample_idx_colpos(self.current_config_key)
        extdict = {".csv": ",", ".tsv": "\t", ".tab": "\t"}
        mapping_fields = self.software_config[self.current_config_key].get("content")

        if not mapping_fields:
            return {}

        if conf_tab_name.endswith(".gz"):
            inner_ext = os.path.splitext(conf_tab_name.strip(".gz"))[1]
            if inner_ext in extdict:
                self.update_all_logs(
                    method_name,
                    "warning",
                    f"Expected tabular file '{conf_tab_name}' is compressed and cannot be processed.",
                )
            return {}

        if file_ext in extdict:
            try:
                return relecov_tools.utils.read_csv_file_return_dict(
                    file_name=file_list[0],
                    sep=extdict[file_ext],
                    key_position=sample_idx_colpos,
                )
            except FileNotFoundError as e:
                self.update_all_logs(
                    method_name,
                    "error",
                    f"Tabular file not found: '{file_list[0]}': {e}",
                )
                raise FileNotFoundError(
                    f"Tabular file not found: '{file_list[0]}'"
                ) from e
        else:
            self.update_all_logs(
                method_name,
                "error",
                f"Unrecognized defined file name extension '{file_ext}' in '{conf_tab_name}'.",
            )
            raise ValueError(self.log_report.print_log_report(method_name, ["error"]))

    def handling_files(self, file_list, sufix, output_dir, file_tag):
        """Handles different file formats to extract data regardless of their structure.
        The goal is to extract the data contained in files specified in ${file_list},
        using either 'standard' handlers defined in this class or pipeline-specific file handlers.
        (inspired from ./metadata_homogenizer.py)

            A file handler method must generate a data structure as follow:
            {
                'SAMPLE1': {
                    'field1': 'value1'
                    'field2': 'value2'
                    'field3': 'value3'
                },
                SAMPLE2': {
                    'field1': 'value1'
                    'field2': 'value2'
                    'field3': 'value3'
                },
                ...
            }
            Note: ensure that 'field1','field2','field3' corresponds with the values
            especified in the 'content' section of each software configuration scope
            (see: conf/bioinfo_config.json).

        Args:
            file_list (list): A list of file path/s to be processed.
            output_dir (str): Path to save output files from imported method if necessary
            file_tag(str): Tag that will be used for output filenames includes batch date (same as download date) and hex.

        Returns:
            data: A dictionary containing bioinfo metadata handled for each sample.
        """
        method_name = f"{self.add_bioinfo_results_metadata.__name__}:{self.handling_files.__name__}"
        splitted_path = os.path.join(output_dir, "analysis_results")
        file_name = self.software_config[self.current_config_key].get("fn")
        # Parsing files
        current_config = self.software_config[self.current_config_key]
        func_name = current_config.get("function")
        if func_name is None:
            data = self.handling_tables(file_list=file_list, conf_tab_name=file_name)
        else:
            if current_config.get("split_by_batch") is True:
                file_extension = current_config.get("fn").rsplit(".", 1)[1]
                base_filename = current_config.get("fn").rsplit(".", 1)[0]
                pattern = re.compile(
                    f"{base_filename}_{sufix}.{re.escape(file_extension)}"
                )
                matching_files = [
                    f for f in os.listdir(splitted_path) if pattern.match(f)
                ]
                full_paths = [  # noqa: F841
                    os.path.join(splitted_path, f) for f in matching_files
                ]
                try:
                    if func_name.startswith("utils/"):
                        utils_name = "relecov_tools.assets.pipeline_utils.utils"
                        func_name = func_name.split("/", 1)[1]
                    else:
                        utils_name = (
                            f"relecov_tools.assets.pipeline_utils.{self.software_name}"
                        )

                    # Dynamically import the function from the specified module
                    import_statement = f"import {utils_name}"
                    exec(import_statement)
                    # Get method name and execute it.
                    data = eval(
                        utils_name
                        + "."
                        + func_name
                        + "(full_paths, file_tag, '"
                        + self.software_name
                        + "', output_dir)"
                    )

                except Exception as e:
                    self.update_all_logs(
                        self.save_merged_files.__name__,
                        "error",
                        f"Error occurred while parsing '{func_name}': {e}.",
                    )
                    sys.exit(self.log_report.print_log_report(method_name, ["error"]))
            else:
                try:
                    # Dynamically import the function from the specified module
                    if func_name.startswith("utils/"):
                        utils_name = "relecov_tools.assets.pipeline_utils.utils"
                        func_name = func_name.split("/", 1)[1]
                    else:
                        utils_name = (
                            f"relecov_tools.assets.pipeline_utils.{self.software_name}"
                        )

                    # Dynamically import the function from the specified module
                    import_statement = f"import {utils_name}"
                    exec(import_statement)
                    # Get method name and execute it.
                    data = eval(
                        utils_name
                        + "."
                        + func_name
                        + "(file_list, file_tag, '"
                        + self.software_name
                        + "', output_dir)"
                    )
                except Exception as e:
                    self.update_all_logs(
                        self.add_bioinfo_results_metadata.__name__,
                        "error",
                        f"Error occurred while parsing '{func_name}': {e}.",
                    )
                    sys.exit(self.log_report.print_log_report(method_name, ["error"]))
        return data

    def add_fixed_values(self, j_data):
        """Add fixed values to j_data as defined in the bioinformatics configuration (definition: "fixed values")

        Args:
            j_data (list(dict{str:str}): A list of dictionaries containing metadata lab (one item per sample).

        Returns:
            j_data: updated j_data with fixxed values added in it.
        """
        method_name = f"{self.add_fixed_values.__name__}"
        try:
            f_values = self.software_config["fixed_values"]
            for row in j_data:
                row["bioinfo_metadata_file"] = self.out_filename
                for field, value in f_values.items():
                    row[field] = value
            self.update_all_logs(method_name, "valid", "Fields added successfully.")
        except KeyError as e:
            self.update_all_logs(
                method_name, "warning", f"Error found while adding fixed values: {e}"
            )
            pass
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data

    def replace_na_value_if_needed(self, field, raw_val):
        """
        Replace 'NA', None or NaN with 'Not Provided [SNOMED:434941000124101]'
        if the field is not required in the schema.
        """
        required_fields = self.bioinfo_schema.get("required", [])
        is_na = (
            raw_val is None
            or (isinstance(raw_val, str) and raw_val.strip().upper() in ["NA", "NONE"])
            or (isinstance(raw_val, float) and np.isnan(raw_val))
        )
        if is_na and field not in required_fields:
            return "Not Provided [SNOMED:434941000124101]"
        return raw_val

    def add_bioinfo_files_path(self, files_found_dict, j_data):
        """Adds file paths essential for handling and mapping bioinformatics metadata to the j_data.
        For each sample in j_data, the function assigns the corresponding file path based on the identified files in files_found_dict.

        If multiple files are identified per configuration item (e.g., viralrecon.mapping_consensus → *.consensus.fa), each sample in j_data receives its respective file path.
        If no file path is located, the function appends "Not Provided [SNOMED:434941000124101]" to indicate missing data.

        Args:
            files_found_dict (dict): A dictionary containing file paths identified for each configuration item.
            j_data (list(dict{str:str}): A list of dictionaries containing metadata lab (one item per sample).

        Returns:
            j_data: Updated j_data with file paths mapped for bioinformatic metadata.
        """
        method_name = f"{self.add_bioinfo_files_path.__name__}"
        sample_name_error = 0
        multiple_sample_files = self.get_multiple_sample_files()
        for row in j_data:
            if not row.get("sequencing_sample_id"):
                self.update_all_logs(
                    method_name,
                    "warning",
                    f'Sequencing_sample_id missing in {row.get("collecting_sample_id")}... Skipping...',
                )
                continue
            sample_name = row["sequencing_sample_id"]
            base_cod_path = row.get("sequence_file_path_R1")
            if base_cod_path is None:
                self.update_all_logs(
                    method_name,
                    "error",
                    f"No 'sequence_file_path_R1' found for sample {sample_name}. Unable to generate paths.",
                )
                continue
            for key, values in files_found_dict.items():
                file_path = []
                if values:  # Check if value is not empty
                    for file in values:
                        if key in multiple_sample_files:
                            file_path.append(file)
                        elif sample_name in file:
                            file_path.append(file)
                else:
                    file_path.append("Not Provided [SNOMED:434941000124101]")
                if self.software_config[key].get("filepath_name"):
                    path_key = self.software_config[key].get("filepath_name")
                    analysis_results_paths = []
                    for paths in file_path:
                        if file_path != "Not Provided [SNOMED:434941000124101]" and (
                            self.software_config[key].get("extract")
                            or self.software_config[key].get("function")
                        ):
                            analysis_results_path = os.path.join(
                                base_cod_path,
                                "analysis_results",
                                os.path.basename(paths),
                            )
                            analysis_results_paths.append(analysis_results_path)
                        else:
                            analysis_results_paths = file_path
                    row[path_key] = ", ".join(analysis_results_paths)
                else:
                    path_key = key
                if self.software_config[key].get("extract"):
                    self.extract_file(
                        file=file_path,
                        dest_folder=row.get("sequence_file_path_R1"),
                        sample_name=sample_name,
                        path_key=path_key,
                    )
        self.log_report.print_log_report(method_name, ["warning"])
        if sample_name_error == 0:
            self.update_all_logs(method_name, "valid", "File paths added successfully.")
        self.log_report.print_log_report(method_name, ["valid", "warning"])
        return j_data

    def collect_info_from_lab_json(self):
        """Reads lab metadata from a JSON file and creates a list of dictionaries.
        Reads lab metadata from the specified JSON file and converts it into a list of dictionaries.
        This list is used to add the rest of the fields.

        Returns:
            json_lab_data: A list of dictionaries containing lab metadata (aka j_data).
        """
        method_name = f"{self.collect_info_from_lab_json.__name__}"
        try:
            json_lab_data = relecov_tools.utils.read_json_file(
                self.readlabmeta_json_file
            )
        except ValueError:
            self.update_all_logs(
                self.collect_info_from_lab_json.__name__,
                "error",
                f"Invalid lab-metadata json file: self.{self.readlabmeta_json_file}",
            )
            sys.exit(self.log_report.print_log_report(method_name, ["error"]))
        return json_lab_data

    def get_sample_idx_colpos(self, config_key):
        """Extract which column contain sample names for that specific file from config

        Returns:
            sample_idx_possition: column number which contains sample names
        """
        try:
            sample_idx_colpos = self.software_config[config_key]["sample_col_idx"] - 1
        except (KeyError, TypeError):
            sample_idx_colpos = 0
            self.update_all_logs(
                "get_sample_idx_colpos",
                "warning",
                f"No valid sample-index-column defined in '{self.software_name}.{config_key}'. Using default instead.",
            )
        return sample_idx_colpos

    def extract_file(self, file, dest_folder, sample_name=None, path_key=None):
        """Copy input file to the given destination, include sample name and key in log
        Args:
            file (str): Path the file that is going to be copied
            dest_folder (str): Folder with files from batch of samples
            sample_name (str, optional): Name of the sample in metadata. Defaults to None.
            path_key (str, optional): Metadata field for the file. Defaults to None.
        Returns:
            bool: True if the process was successful, else False
        """
        dest_folder = os.path.join(dest_folder, "analysis_results")
        os.makedirs(dest_folder, exist_ok=True)
        for filepath in file:
            out_filepath = os.path.join(dest_folder, os.path.basename(filepath))
            if os.path.isfile(out_filepath):
                self.log.debug(f"{out_filepath} already exists, not extracted")
                continue
            if filepath == "Not Provided [SNOMED:434941000124101]":
                self.update_all_logs(
                    self.extract_file.__name__,
                    "warning",
                    f"File for {path_key} not provided in sample {sample_name}",
                )
                continue
            try:
                shutil.copy(filepath, out_filepath)
            except (IOError, PermissionError) as e:
                self.update_all_logs(
                    self.extract_file.__name__,
                    "warning",
                    f"Could not extract {filepath}: {e}",
                )
                continue
        return True

    def split_data_by_batch(self, j_data):
        """Split metadata from json for each batch of samples found according to folder location of the samples.
        Args:
            files_found_dict (dict): A dictionary containing file paths identified for each configuration item.
            j_data (list(dict)): List of dictionaries, one per sample, including metadata for that sample
        Returns:
            data_by_batch (dict(list(dict))): Dictionary containing parts of j_data corresponding to each
            different folder with samples (batch) included in the original json metadata used as input
        """
        unique_batchs = set([x.get("sequence_file_path_R1") for x in j_data])
        data_by_batch = {batch_dir: {} for batch_dir in unique_batchs}
        for batch_dir in data_by_batch.keys():
            data_by_batch[batch_dir]["j_data"] = [
                samp
                for samp in j_data
                if samp.get("sequence_file_path_R1") == batch_dir
            ]
        return data_by_batch

    def split_tables_by_batch(self, files_found_dict, sufix, batch_data, output_dir):
        """Filter table content to output a new table containing only the samples present in given metadata
        Args:
            files_found_dict (dict): A dictionary containing file paths identified for each configuration item.
            sufix (str): Sufix to be added to the new table file name.
            batch_data (list(dict)): Metadata corresponding to a single folder with samples (folder)
            output_dir (str): Output location for the generated tabular file
        """

        def extract_batch_rows_to_file(file, sufix):
            """Create a new table file only with rows matching samples in batch_data"""
            extdict = {".csv": ",", ".tsv": "\t", ".tab": "\t"}
            file_extension = os.path.splitext(file)[1]
            file_df = pd.read_csv(
                file, sep=extdict.get(file_extension), header=header_pos
            )
            sample_col = file_df.columns[sample_colpos]
            file_df[sample_col] = file_df[sample_col].astype(str)
            file_df = file_df[file_df[sample_col].isin(batch_samples)]

            base, ext = os.path.splitext(os.path.basename(file))
            new_filename = f"{base}_{sufix}{ext}"
            os.makedirs(os.path.join(output_dir, "analysis_results"), exist_ok=True)
            output_path = os.path.join(output_dir, "analysis_results", new_filename)
            file_df.to_csv(output_path, index=False, sep=extdict.get(file_extension))
            return

        method_name = self.split_tables_by_batch.__name__
        namekey = "sequencing_sample_id"
        batch_samples = [row.get(namekey) for row in batch_data]
        for key, files in files_found_dict.items():
            if not self.software_config[key].get("split_by_batch"):
                continue
            header_pos = self.software_config[key].get("header_row_idx", 1) - 1
            sample_colpos = self.get_sample_idx_colpos(key)
            for file in files:
                try:
                    extract_batch_rows_to_file(file, sufix)
                except Exception as e:
                    if self.software_config[key].get("required"):
                        log_type = "error"
                    else:
                        log_type = "warning"
                    self.update_all_logs(
                        method_name,
                        log_type,
                        f"Could not create batch table for {file}: {e}",
                    )
        self.log_report.print_log_report(method_name, ["valid", "warning", "error"])
        return

    def merge_metadata(self, batch_filepath, batch_data):
        """
        Merge metadata json if sample does not exist in the metadata file,
        or prompt the user to update if --update flag is provided and sample differs.

        Args:
            batch_filepath (str): Path to save the json file with the metadata.
            batch_data (list): A list of dictionaries containing metadata of the samples.

        Returns:
            merged_metadata (list): The updated list of metadata entries.
        """
        merged_metadata = relecov_tools.utils.read_json_file(batch_filepath)
        prev_metadata_dict = {
            item["sequencing_sample_id"]: item for item in merged_metadata
        }

        overwrite_all = False

        for item in batch_data:
            sample_id = item["sequencing_sample_id"]
            if sample_id in prev_metadata_dict:
                if prev_metadata_dict[sample_id] != item:
                    if self.update:
                        if not overwrite_all:
                            stderr.print(
                                f"[red]Sample '{sample_id}' has different metadata than the existing entry in {batch_filepath}."
                            )
                            response = Prompt.ask(
                                "[yellow]Do you want to overwrite this sample?",
                                choices=["y", "n", "all"],
                                default="n",
                            )
                            if response == "all":
                                overwrite_all = True
                                stderr.print(
                                    "[green]All subsequent samples will be overwritten automatically."
                                )
                                prev_metadata_dict[sample_id] = item
                            elif response == "y":
                                prev_metadata_dict[sample_id] = item
                                stderr.print(f"[green]Sample '{sample_id}' updated.")
                            else:
                                stderr.print(
                                    f"[blue]Skipping update for sample '{sample_id}'."
                                )
                        else:
                            prev_metadata_dict[sample_id] = item
                            stderr.print(f"[green]Sample '{sample_id}' updated (auto).")
                    else:
                        stderr.print(
                            f"[red]Sample '{sample_id}' has different data in {batch_filepath} and new metadata. Can't merge."
                        )
                        self.log.error(
                            "Sample %s has different data in %s and new metadata. Can't merge.",
                            sample_id,
                            batch_filepath,
                        )
                        sys.exit(1)
            else:
                prev_metadata_dict[sample_id] = item

        merged_metadata = list(prev_metadata_dict.values())
        relecov_tools.utils.write_json_to_file(merged_metadata, batch_filepath)
        return merged_metadata

    def save_merged_files(self, files_dict, file_tag, output_dir=None):
        """
        Process and save files that where split by cod and that have a function to be processed

        Args:
            files_dict (dict): A dictionary containing file paths identified for each configuration item.
            file_tag (str): Tag that will be used for output filenames includes batch date (same as download date) and hex.
            output_dir (str): Path to save output files generated during processing.

        Returns:
            None
        """
        method_name = f"{self.save_merged_files.__name__}"
        for key, config in self.software_config.items():
            func_name = config.get("function")
            # Skip configurations that do not match the conditions
            if config.get("split_by_batch") is True and func_name:
                try:
                    file_path = files_dict[key]
                    stderr.print(f"[blue]Processing splitted file: {file_path}")
                    self.log.info(f"Processing splitted file: {file_path}")
                except KeyError:
                    self.update_all_logs(
                        method_name,
                        "warning",
                        f"No file path found for '{file_path}'",
                    )
                    continue
                try:
                    # Dynamically import the function from the specified module
                    if func_name.startswith("utils/"):
                        utils_name = "relecov_tools.assets.pipeline_utils.utils"
                        func_name = func_name.split("/", 1)[1]
                    else:
                        utils_name = (
                            f"relecov_tools.assets.pipeline_utils.{self.software_name}"
                        )

                    import_statement = f"import {utils_name}"
                    exec(import_statement)
                    # Get method name and execute it.
                    data = eval(
                        utils_name
                        + "."
                        + func_name
                        + "(file_path, file_tag, '"
                        + self.software_name
                        + "', output_dir)"
                    )
                except Exception as e:
                    self.update_all_logs(
                        self.save_merged_files.__name__,
                        "error",
                        f"Error occurred while parsing '{func_name}': {e}.",
                    )
                    sys.exit(self.log_report.print_log_report(method_name, ["error"]))
        return data

    def get_multiple_sample_files(self):
        multiple_sample_files = []
        for key in self.software_config.keys():
            if self.software_config[key].get("multiple_samples"):
                multiple_sample_files.append(key)
        return multiple_sample_files

    def filter_properties(self, data):
        """
        Remove properties from bioinfo_metadata that are not in the schema properties.

        Parameters
        ----------
        data : list[dict]
            Dictionary with the bioinfo metadata

        Returns
        -------
        data: list[dict]
            Bioinfo metadata json filtered
        """
        valid_keys = set(self.json_schema.get("properties", {}).keys())

        for sample in data:
            for k in list(sample.keys()):
                if k not in valid_keys:
                    sample.pop(k)
        return data

    def create_bioinfo_file(self):
        """Create the bioinfodata json with collecting information from lab
        metadata json, mapping_stats, and more information from the files
        inside input directory.

        Returns:
            bool: True if the bioinfo file creation process was successful.
        """

        # Check samplesheet for matching samples
        self.validate_samplenames()

        # Find and validate bioinfo files
        stderr.print("[blue]Scanning input directory...")
        files_found_dict = self.scann_directory()
        stderr.print("[blue]Validating required files...")
        self.validate_software_mandatory_files(files_found_dict)
        # Split files found based on each batch of samples
        data_by_batch = self.split_data_by_batch(self.j_data)
        sufix = datetime.now().strftime("%Y%m%d%H%M%S")

        # Add bioinfo metadata to j_data
        for batch_dir, batch_dict in data_by_batch.items():
            batch_data = batch_dict["j_data"]
            if not batch_data:
                self.log.warning(
                    f"Data from batch {batch_dir} was completely empty. Skipped."
                )
                self.update_all_logs(
                    self.create_bioinfo_file.__name__,
                    "warning",
                    f"Data from batch {batch_dir} was completely empty. Skipped.",
                )
                continue
            first_sample = batch_data[0]
            lab_code = first_sample.get(
                "submitting_institution_id", batch_dir.split("/")[-2]
            )
            batch_date = first_sample.get("batch_id", batch_dir.split("/")[-1])
            self.set_batch_id(batch_date)
            self.logsum.feed_key(batch_dir)
            file_tag = batch_date + "_" + self.hex
            stderr.print(f"[blue]Processing data from {batch_dir}")
            stderr.print("[blue]Adding bioinfo metadata to read lab metadata...")
            self.split_tables_by_batch(files_found_dict, sufix, batch_data, batch_dir)
            batch_data = self.add_bioinfo_results_metadata(
                files_found_dict, batch_data, sufix, file_tag, batch_dir
            )
            stderr.print("[blue]Adding software versions to read lab metadata...")
            stderr.print("[blue]Adding fixed values")
            batch_data = self.add_fixed_values(batch_data)
            # Adding files path
            stderr.print("[blue]Adding files path to read lab metadata")
            batch_data = self.add_bioinfo_files_path(files_found_dict, batch_data)
            tag = "bioinfo_lab_metadata_"
            batch_filename = tag + lab_code + ".json"
            batch_filename = self.tag_filename(batch_filename)
            batch_filepath = os.path.join(batch_dir, batch_filename)
            module = eval(f"relecov_tools.assets.pipeline_utils.{self.software_name}")
            try:
                if hasattr(module, "quality_control_evaluation"):
                    qc_func = eval(
                        f"relecov_tools.assets.pipeline_utils.{self.software_name}.quality_control_evaluation"
                    )
                    qc_data = qc_func(batch_data)
                    for sample in batch_data:
                        sample_id = sample.get("sequencing_sample_id")
                        if sample_id in qc_data:
                            sample.update(qc_data[sample_id])
            except (AttributeError, NameError, TypeError, ValueError) as e:
                self.update_all_logs(
                    self.create_bioinfo_file.__name__,
                    "warning",
                    f"Could not evaluate quality_control_evaluation for batch {batch_dir}: {e}",
                )
                stderr.print(
                    f"[orange]Could not evaluate quality_control_evaluation for batch {batch_dir}: {e}"
                )

            # Filter properties from batch_data that are not included in the schema
            batch_data = self.filter_properties(batch_data)

            if os.path.exists(batch_filepath):
                stderr.print(
                    f"[blue]Bioinfo metadata {batch_filepath} file already exists. Merging new data if possible."
                )
                self.log.info(
                    "Bioinfo metadata %s file already exists. Merging new data if possible."
                    % batch_filepath
                )
                batch_data = self.merge_metadata(batch_filepath, batch_data)
            else:
                relecov_tools.utils.write_json_to_file(batch_data, batch_filepath)
            for sample in batch_data:
                self.logsum.feed_key(
                    key=batch_dir, sample=sample.get("sequencing_sample_id")
                )
            self.log.info("Created output json file: %s" % batch_filepath)
            stderr.print(f"[green]Created batch json file: {batch_filepath}")

        year = str(datetime.now().year)
        out_path = os.path.join(self.output_dir, year)
        os.makedirs(out_path, exist_ok=True)

        stderr.print("[blue]Saving previously splitted files to output directory")
        batch_dates = []
        # Get batch date for all the samples
        for batch_dir, batch_dict in data_by_batch.items():
            if batch_dir.split("/")[-1] not in batch_dates:
                batch_dates.append(batch_dir.split("/")[-1])

        if len(batch_dates) == 1:
            batch_date = str(batch_dates[0])
        else:
            stderr.print(
                "[orange]More than one batch date in the same json data. Using current date as batch date."
            )
            self.log.info(
                "More than one batch date in the same json data. Using current date as batch date."
            )
            batch_date = self.basemod_date
        self.set_batch_id(batch_date)
        file_tag = self.batch_id + "_" + self.hex
        self.save_merged_files(files_found_dict, file_tag, out_path)
        batch_filename = self.tag_filename("bioinfo_lab_metadata" + ".json")
        stderr.print("[blue]Writting output json file")
        file_path = os.path.join(out_path, batch_filename)
        if os.path.exists(file_path):
            stderr.print(
                f"[blue]Bioinfo metadata {file_path} file already exists. Merging new data if possible."
            )
            self.log.info(
                "Bioinfo metadata %s file already exists. Merging new data if possible."
                % file_path
            )
            batch_data = self.merge_metadata(file_path, self.j_data)
        else:
            relecov_tools.utils.write_json_to_file(self.j_data, file_path)
        stderr.print(f"[green]Sucessful creation of bioinfo analyis file: {file_path}")
        self.parent_create_error_summary(
            called_module="read-bioinfo-metadata", logs=self.logsum.logs
        )
        return True



#!/usr/bin/env python
from collections import OrderedDict
from datetime import datetime
import json
import jsonschema
from jsonschema import Draft202012Validator
from relecov_tools.config_json import ConfigJson
import rich.console
import os
import sys

# import jsonschema
import relecov_tools.utils
from relecov_tools.base_module import BaseModule

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class MappingSchema(BaseModule):
    def __init__(
        self,
        origin_schema=None,
        json_file=None,
        destination_schema=None,
        schema_file=None,
        output_dir=None,
    ):
        super().__init__(output_dir=output_dir, called_module=__name__)
        config_json = ConfigJson()
        self.config_json = config_json
        if origin_schema is None:
            origin_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "schema",
                config_json.get_topic_data("json_schemas", "relecov_schema"),
            )
        else:
            if not os.path.isfile(origin_schema):
                self.log.error("Relecov schema file %s does not exist", origin_schema)
                stderr.print(
                    "[red] Relecov schema " + origin_schema + " does not exist"
                )
                exit(1)
        rel_schema_json = relecov_tools.utils.read_json_file(origin_schema)
        try:
            Draft202012Validator.check_schema(rel_schema_json)
        except jsonschema.ValidationError:
            self.log.error("Relecov schema does not fulfill Draft 202012 Validation ")
            stderr.print(
                "[red] Relecov schema does not fulfill Draft 202012 Validation"
            )
            sys.exit(1)
        self.relecov_schema = rel_schema_json

        if json_file is None:
            json_file = relecov_tools.utils.prompt_path(
                msg="Select the json file which have the data to map"
            )
        if not os.path.isfile(json_file):
            self.log.error("json data file %s does not exist ", json_file)
            stderr.print(f"[red] json data file {json_file} does not exist")
            sys.exit(1)
        self.json_data = relecov_tools.utils.read_json_file(json_file)
        self.json_file = json_file
        batch_id = self.get_batch_id_from_data(self.json_data)
        self.set_batch_id(batch_id)

        if destination_schema is None:
            self.destination_schema = relecov_tools.utils.prompt_selection(
                msg="Select ENA, GISAID for already defined schemas or other for custom",
                choices=["ENA", "GISAID", "other"],
            )
        else:
            self.destination_schema = destination_schema
        if self.destination_schema == "other":
            if schema_file is None:
                self.schema_file = relecov_tools.utils.prompt_path(
                    msg="Select the json schema file to map your data"
                )
            else:
                self.schema_file = schema_file
            if not os.path.exists(self.schema_file):
                self.log.error(
                    "Schema file %s to map your data does not exist ",
                    self.metadata_file,
                )
                sys.exit(1)
            with open(self.schema_file, "r") as fh:
                json_schema = json.load(fh)
            try:
                Draft202012Validator.check_schema(json_schema)
            except jsonschema.ValidationError:
                stderr.print(
                    "[red] Json schema does not fulfill Draft 202012 Validation"
                )
                sys.exit(1)
        elif self.destination_schema == "ENA":
            self.schema_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "schema",
                config_json.get_topic_data("json_schemas", "ena_schema"),
            )
        elif self.destination_schema == "GISAID":
            self.schema_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "schema",
                config_json.get_topic_data("json_schemas", "gisaid_schema"),
            )
        else:
            stderr.print("[red] Invalid option for mapping to schena")
            sys.exit(1)
        with open(self.schema_file, "r") as fh:
            self.mapped_to_schema = json.load(fh)

        self.ontology = {}
        for key, values in self.relecov_schema["properties"].items():
            if values["ontology"] == "0":
                continue
            self.ontology[values["ontology"]] = key
        self.output_dir = output_dir

        if os.path.exists(os.path.join(output_dir, "mapping_errors.log")):
            os.remove(os.path.join(output_dir, "mapping_errors.log"))

    def maping_schemas_based_on_geontology(self):
        """Return a dictionary with the properties of the mapped_to_schema as key and
        properties of Relecov Schema as value
        """
        mapped_dict = OrderedDict()
        errors = {}
        required_fields = self.mapped_to_schema["required"]

        for key, values in self.mapped_to_schema["properties"].items():
            if values["ontology"] == "0":
                continue
            try:
                mapped_dict[key] = self.ontology[values["ontology"]]
            except KeyError as e:
                if key in required_fields:
                    stderr.print(
                        f"[red]Required field {key} ontology missing in relecov schema"
                    )
                    sys.exit(1)
                else:
                    errors[key] = str(e)
        if len(errors) >= 1:
            output_errs = "\n".join(f"{field}:{info}" for field, info in errors.items())
            invalid_ontologies = str([field for field in errors.keys()]).strip("[]")
            self.log.error("Invalid ontology for: " + invalid_ontologies)
            stderr.print("[yellow]\nGot unmapped ontologies. Check mapping_errors.log")
            with open("mapping_errors.log", "w") as errs:
                errs.write("Ontology mapping errors:\n" + output_errs + "\n")
        return mapped_dict

    def mapping_json_data(self, mapping_schema_dict):
        """Convert phage plus data to the requested schema"""
        mapped_data = []

        for data in self.json_data:
            map_sample_dict = OrderedDict()
            for item, value in mapping_schema_dict.items():
                try:
                    if isinstance(data[value], str):
                        data[value] = data[value].split(" [", 1)[0]
                    map_sample_dict[item] = data[value]
                except KeyError as e:
                    self.log.info("Property %s not set in the source data", e)
            mapped_data.append(map_sample_dict)
        return mapped_data

    def additional_formating(self, mapped_json_data):
        """Update data that needs additional formating such as
        word splitting and include fields with fixed values.
        """
        additional_data = self.config_json.get_topic_data(
            "ENA_fields", "additional_formating"
        )
        fixed_fields = self.config_json.get_topic_data("ENA_fields", "ena_fixed_fields")

        if self.destination_schema == "ENA":
            for idx in range(len(self.json_data)):
                for key, value in fixed_fields.items():
                    mapped_json_data[idx][key] = value
                for key, _ in additional_data.items():
                    """
                    Some fields in ENA need special formatting such as sample_id+date.
                    Instead of directly merging them, -- is used as delimiter.
                    Also, the Not Provided fields are skipped in this process
                    """
                    formated_data = {
                        x: "--".join(
                            [
                                self.json_data[idx].get(f, "").split(" [", 1)[0]
                                for f in y
                                if "Not Provided"
                                not in self.json_data[idx].get(f, "").split(" [", 1)[0]
                            ]
                        )
                        for x, y in additional_data.items()
                    }
                    if "fastq_filepath" in key:
                        formated_data[key] = formated_data[key].replace("--", "/")
                    mapped_json_data[idx][key] = formated_data[key]
        elif self.destination_schema == "GISAID":
            for idx in range(len(self.json_data)):
                mapped_json_data[idx]["covv_type"] = "betacoronavirus"
        """
        This is a temporal solution for library_strategy. Once the values are also
        mapped by the ontology (not only the fields) this should not be necessary
        """
        for sample in mapped_json_data:
            if not sample.get("library_strategy"):
                continue
            sample["library_strategy"] = sample["library_strategy"].strip(" strategy")

        return mapped_json_data

    def check_required_fields(self, mapped_json_data, dest_schema):
        """Checks which required fields are Not Provided"""
        if dest_schema == "ENA":
            # The block below can probably go into an auxiliar function
            required_fields = self.mapped_to_schema["required"]
            for sample in mapped_json_data:
                missing_required = [x for x in required_fields if x not in sample]
                for field in missing_required:
                    sample[field] = "Not Provided"
            try:
                not_provided_fields = {
                    sample["isolate"]: [
                        field
                        for field in required_fields
                        if "Not Provided" in sample[field]
                    ]
                    for sample in mapped_json_data
                }
            except KeyError as e:
                print(f"Field {e} could not be found in json data. Aborting")
                print()
                sys.exit(1)
            notprov_report = "\n".join(
                f"Sample {key}: {str(val).strip('[]')}"
                for key, val in not_provided_fields.items()
            )
            self.log.error(
                f"Some required fields for {dest_schema} were Not Provided: "
                "Check mapping_errors.log for more details"
            )
            stderr.print(
                f"[yellow]\nSome required fields for {dest_schema} were Not Provided:",
                "[yellow]\nCheck mapping_errors.log for more details",
            )
            self.log.error("Required fields Not Provided:\n" + notprov_report)
            with open("mapping_errors.log", "a") as errs:
                errs.write("Required fields Not Provided:\n" + notprov_report)
        else:
            # Only ENA's schema is supported as yet
            return
        return mapped_json_data

    def write_json_fo_file(self, mapped_json_data):
        """Write metadata to json file"""
        os.makedirs(self.output_dir, exist_ok=True)
        time = datetime.now().strftime("%Y_%m_%d")
        file_name = (
            "mapped_metadata" + "_" + self.destination_schema + "_" + time + ".json"
        )
        json_file = os.path.join(self.output_dir, file_name)
        stderr.print("Writting mapped data to json file:", json_file)
        with open(json_file, "w", encoding="utf-8") as fh:
            fh.write(
                json.dumps(
                    mapped_json_data, indent=4, sort_keys=True, ensure_ascii=False
                )
            )
        return True

    def map_to_data_to_new_schema(self):
        """Mapping the json data from relecov schema to the requested one"""
        mapping_schema_dict = self.maping_schemas_based_on_geontology()
        mapped_json_data = self.mapping_json_data(mapping_schema_dict)
        updated_json_data = self.additional_formating(mapped_json_data)
        self.check_required_fields(mapped_json_data, self.destination_schema)
        self.write_json_fo_file(updated_json_data)
        stderr.print(f"[green]Finished mapping to {self.destination_schema} schema")
        return



#!/usr/bin/env python
import sys

import logging
import rich.console
import relecov_tools.utils

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


def replace_originating_lab(metadata, f_data, mapped_fields, heading):
    """Replace the format text in the originating lab and replace by the ones
    defined in the System
    """
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            meta_idx = heading.index(key)
            try:
                row[meta_idx] = f_data[row[meta_idx].strip()][val].strip()
            except KeyError as e:
                log.error("Value  %s does not exist ", e)
                stderr.print(f"[red] Value {e} does not exist")
                sys.exit(1)
    return metadata


def added_seq_inst_model(metadata, f_data, mapped_fields, heading):
    """Set the type of sequencer instrument based on the run name"""
    s_idx = heading.index("Sample ID given for sequencing")
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            try:
                run_name = f_data[str(row[s_idx])][val].lower()
            except KeyError as e:
                log.error("Value  %s does not exist ", e)
                stderr.print(f"[red] Value {e} does not exist")
                sys.exit(1)
            if "nextseq" in run_name:
                row[m_idx] = "Illumina NextSeq 500"
            elif "next_seq" in run_name:
                row[m_idx] = "Illumina NextSeq 500"
            elif "miseq" in run_name:
                row[m_idx] = "Illumina MiSeq"
            elif "miseaq" in run_name:
                row[m_idx] = "Illumina MiSeq"
            elif "novaseq" in run_name:
                row[m_idx] = "Illumina NovaSeq 6000"
            else:
                log.error("Value  %s is not defined in the mapping ", run_name)
                stderr.print(f"[red] Value {run_name} is not defined in the mapping")
                sys.exit(1)
    return metadata


def translate_gender_to_english(metadata, f_data, mapped_fields, heading):
    """Translate into english the host gender that is written in spanish"""
    map_dict = {
        "hombre": "Male",
        "mujer": "Female",
        "genero no-binario": "Non-binary Gender",
        "genero no-binario": "Non-binary Gender",
        "desconocido": "Not Provided",
        "unknown": "Not Provided",
    }
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            if row[m_idx] is None or row[m_idx] == "":
                row[m_idx] = "Not Provided"
                continue
            item = row[m_idx].lower()
            if item in map_dict:
                row[m_idx] = map_dict[item]
            else:
                log.error("The '%s' is not a valid data for translation", row[m_idx])
                stderr.print(
                    "f[red] The '{row[m_idx]}' is not a valid data for translation"
                )
                sys.exit(1)
    return metadata


def translate_specimen_source(metadata, f_data, mapped_fields, heading):
    """Translate into english the "muestra" that is written in spanish"""
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            if row[m_idx] is None:
                row[m_idx] = "not provided"
            elif "ASPIRADO NASOFARÍNGEO" in row[m_idx].upper():
                row[m_idx] = "Nasopharynx Aspiration"
            elif "ASPIRADO BRONQUIAL" in row[m_idx].upper():
                row[m_idx] = "Bronchus Aspiration"
            elif "ESPUTO" in row[m_idx].upper():
                row[m_idx] = "Sputum"
            elif "EXTRACTO" in row[m_idx].upper():
                row[m_idx] = "Scraping"
            elif "EXUDADO FARÍNGEO" in row[m_idx].upper():
                row[m_idx] = "Pharynx Swabbing"
            elif "EXUDADO NASOFARÍNGEO" in row[m_idx].upper():
                row[m_idx] = "Nasopharynx Swabbing"
            elif "EXUDADO OROFARINGEO" in row[m_idx].upper():
                row[m_idx] = "Oropharynx Swabbing"
            elif "PLACENTA" in row[m_idx].upper():
                row[m_idx] = "Placenta"
            elif "SALIVA" in row[m_idx].upper():
                row[m_idx] = "Saliva"
            else:
                log.error("The field is not correctly written or is not filled")
                stderr.print("The field is not correctly written or not filled")
                sys.exit(1)
    return metadata


def translate_purpose_seq_to_english(metadata, f_data, mapped_fields, heading):
    """Fetch the first words of the option to group them according the
    schema
    """
    map_dict = {
        "estudio variante": "Targeted surveillance (non-random sampling)",
        "trabajador/a granja visones": "Targeted surveillance (non-random sampling)",
        "sospecha reinfección": "Re-infection surveillance",
        "i-move-covid": "Research",
        "irag": "Research",
        "muestreo aleatorio": "Baseline surveillance (random sampling)",
        "paciente vacunado": "Vaccine escape surveillance",
        "posible variante": "Sample has epidemiological link to Variant of Concern (VoC)",
        "no consta": "Not Collected",
        "brote": "Cluster/Outbreak investigation",
        "viaje": "Surveillance of international border crossing by air travel or ground transport",
        "posible variante": "Sample has epidemiological link to Variant of Concern (VoC)",
    }
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            if row[m_idx] is None or row[m_idx] == "":
                row[m_idx] = "Not Provided"
                continue
            item = row[m_idx].lower()
            if item in map_dict:
                row[m_idx] = map_dict[item]
            elif "brote" in item:
                row[m_idx] = map_dict["brote"]
            elif "viaje" in item:
                row[m_idx] = map_dict["viaje"]
            elif "posible variante" in item:
                row[m_idx] = map_dict["posible variante"]
            else:
                log.error("The '%s' is not a valid data for translation", row[m_idx])
                stderr.print(
                    "f[red] The {row[m_idx]} is not a valid data for translation"
                )
                sys.exit(1)
    return metadata


def translate_nucleic_acid_extract_prot(metadata, f_data, mapped_fields, heading):
    """Fetch the short name given in the input laboratory file and change for
    the one is allow according to schema
    """
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            if "NA" in row[m_idx]:
                row[m_idx] = "Not Applicable"
            elif "opentrons" in row[m_idx].lower():
                row[m_idx] = "Opentrons custom rna extraction protocol"
            else:
                # allow from now on until more options are available
                continue
    return metadata


def findout_library_layout(metadata, f_data, mapped_fields, heading):
    """Read the file and by checking if read2_cycles is 0 set to Single otherwise
    to paired"""
    s_idx = heading.index("Sample ID given for sequencing")
    for row in metadata[1:]:
        for key, val in mapped_fields.items():
            m_idx = heading.index(key)
            try:
                if f_data[str(row[s_idx])][val] == "0":
                    row[m_idx] = "Single"
                else:
                    row[m_idx] = "Paired"
            except KeyError as e:
                log.error("The %s is not defined in function findout_library_layout", e)
                stderr.print(
                    f"[red] {e} is not defined in function findout_library_layout"
                )
                sys.exit(1)
    return metadata



import pandas as pd
import json

csv_file = "Hospitales_de_Espaadb.csv"
df = pd.read_csv(csv_file)

json_dict = {}


def to_title_case(value, default="Desconocido"):
    if pd.isna(value) or value == "":
        return default
    return str(value).strip().title()


for index, row in df.iterrows():
    hospital_name = to_title_case(row["NOMBRE"])
    json_dict[hospital_name] = {
        "collecting_institution_address": to_title_case(row["DIRECCION"]),
        "collecting_institution_email": (
            row["EMAIL"].strip() if not pd.isna(row["EMAIL"]) else "Desconocido"
        ),
        "collecting_institution_finalidad": to_title_case(
            row["FINALIDAD_ASISITENCIAL"]
        ),
        "geo_loc_state": to_title_case(row["COMUNIDADES"]),
        "geo_loc_region": to_title_case(row.get("PROVINCIAS", None)),
        "geo_loc_city": to_title_case(row.get("MUNICIPIOS", None)),
        "geo_loc_country": to_title_case(row.get("pais", None), default="Spain"),
        "submitting_institution": "",
        "submitting_institution_address": "",
        "submitting_institution_email": "",
    }

output_file = "laboratory_address.json"
with open(output_file, "w", encoding="utf-8") as json_file:
    json.dump(json_dict, json_file, ensure_ascii=False, indent=4)

print(f"Archivo JSON generado: {output_file}")

json_dict_coords = {}


for index, row in df.iterrows():
    hospital_name = to_title_case(row["NOMBRE"])
    json_dict_coords[hospital_name] = {
        "geo_loc_latitude": f"{row['Y']:.4f}",
        "geo_loc_longitude": f"{row['X']:.4f}",
    }
output_file_coords = "geo_loc_centers.json"
with open(output_file_coords, "w", encoding="utf-8") as json_file:
    json.dump(json_dict_coords, json_file, ensure_ascii=False, indent=4)

print(f"Archivo JSON generado: {output_file_coords}")



#!/usr/bin/env python
import logging
import json
import requests
import rich.console
import relecov_tools.utils

log = logging.getLogger(__name__)

stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class RestApi:
    def __init__(self, server, url):
        self.request_url = server + url
        self.headers = {"content-type": "application/json"}

    def get_request(self, request_info, parameter, value=None, safe=True):
        if parameter == "" or parameter is None:
            url_http = str(self.request_url + request_info)
        elif isinstance(parameter, dict):
            param_value = []
            for key, value in parameter.items():
                param_value.append(key + "=" + value)
            url_http = str(
                self.request_url + request_info + "?" + "&".join(param_value)
            )
        else:
            url_http = str(
                self.request_url + request_info + "?" + parameter + "=" + value
            )
        try:
            req = requests.get(url_http, headers=self.headers)
            if req.status_code != 200:
                if safe:
                    log.error(
                        "Unable to get parameters. Received error code %s",
                        req.status_code,
                    )
                    stderr.print(
                        "[red] Unable to fetch data. Received error ", req.status_code
                    )
                return {"ERROR": req.status_code}
            return {"DATA": json.loads(req.text)}
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return {"ERROR": "Server not available"}

    def put_request(self, data, credentials, url):
        if isinstance(credentials, dict):
            auth = (credentials["user"], credentials["pass"])
        url_http = str(self.request_url + url)
        try:
            req = requests.put(url_http, data=data, auth=auth)
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return {"ERROR": "Server not available"}
        if req.status_code != 201:
            log.error(
                "Unable to post parameters. Received error code %s",
                req.status_code,
            )
            if req.status_code != 500:
                stderr.print(f"[red] Unable to put data because  {req.text}")
            stderr.print(f"[red] Received error {req.status_code}")
            return {"ERROR": req.status_code}
        return {"Success": req.text}

    def post_request(self, data, credentials, url, file=None):
        if isinstance(credentials, dict):
            auth = (credentials["user"], credentials["pass"])
        url_http = str(self.request_url + url)
        try:
            if file:
                files = {"upload_file": open(file, "rb")}
                req = requests.post(
                    url_http, files=files, data=data, headers=self.headers, auth=auth
                )
            else:
                req = requests.post(
                    url_http, data=data, headers=self.headers, auth=auth
                )
            if req.status_code != 201:
                log.error(
                    "Unable to post parameters. Received error code %s",
                    req.status_code,
                )
                stderr.print(f"[red] Received error {req.status_code}")
                if req.status_code != 500:
                    stderr.print(f"[red] Unable to post data because  {req.text}")
                    return {"ERROR": req.status_code, "ERROR_TEST": req.text}
                else:
                    return {"ERROR": req.status_code, "ERROR_TEST": ""}
            return {"Success": req.text}
        except requests.ConnectionError:
            log.error("Unable to open connection towards %s", self.request_url)
            stderr.print("[red] Unable to open connection towards ", self.request_url)
            return {"ERROR": "Server not available"}



#!/usr/bin/env python
import os
import json
import relecov_tools.utils
import relecov_tools.sftp_client
import relecov_tools.mail
from secrets import token_hex
from datetime import datetime
from rich.console import Console
from relecov_tools.config_json import ConfigJson
from relecov_tools.base_module import BaseModule
import subprocess

stderr = Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class UploadSftp(BaseModule):
    def __init__(
        self,
        user=None,
        password=None,
        batch_id=None,
        template_path=None,
        project="Relecov",
    ):
        """Starts the SFTP upload process"""
        super().__init__(called_module=__name__)
        self.log.info(f"Beginning upload process for batch: {batch_id}")

        if not batch_id:
            raise ValueError("Error: You must provide a batch_id as an argument.")
        self.set_batch_id(batch_id)
        config_json = ConfigJson(extra_config=True)
        config = config_json.get_configuration("mail_sender")
        sftp_config = config_json.get_configuration("sftp_handle")
        self.allowed_file_ext = config_json.get_topic_data(
            "sftp_handle", "allowed_file_extensions"
        )

        if not template_path:
            template_path = config.get("delivery_template_path_file")

        if not template_path or not os.path.exists(template_path):
            errtxt = (
                "The template path could not be determined or does not exist. "
                "Please provide it via --template_path or define 'delivery_template_path_file' in the configuration."
            )
            self.log.error(errtxt)
            raise FileNotFoundError(errtxt)

        self.batch_id = batch_id
        self.sftp_user = user or relecov_tools.utils.prompt_text(
            msg="Enter the user id:"
        )
        self.sftp_passwd = password or relecov_tools.utils.prompt_password(
            msg="Enter your password:"
        )

        self.relecov_sftp = relecov_tools.sftp_client.SftpRelecov(
            username=self.sftp_user, password=self.sftp_passwd
        )
        self.log.info(f"User: {self.sftp_user}, Processing batch: {self.batch_id}")
        stderr.print(f"User: {self.sftp_user}, Processing batch: {self.batch_id}")

        self.processed_batches = {}  # Dictionary to store results
        self.template_path = template_path
        self.email_sender = relecov_tools.mail.EmailSender(config, template_path)
        self.guide = config.get("institutions_guide_path")
        self.analysis_folder = sftp_config.get("analysis_results_folder")
        self.project = project
        valid_projects = ["Relecov", "Redlabra"]
        if self.project not in valid_projects:
            self.log.error(f"Error: Only valid projects: {valid_projects}")
            raise ValueError(f"Error: Only valid projects: {valid_projects}")

    def find_cod_for_batch(self):
        """Find all COD* folders containing the batch_id"""
        base_dir = os.getcwd()  # This module should be run in root directory
        cod_dirs = [
            d for d in os.listdir(base_dir) if os.path.isdir(d) and d.startswith("COD")
        ]
        self.log.debug(f"List of directories found: {cod_dirs}")
        matching_cod = {}

        for cod in cod_dirs:
            batch_path = os.path.join(base_dir, cod, self.batch_id)
            if os.path.isdir(batch_path):
                matching_cod[cod] = {"batch": self.batch_id, "path": batch_path}

        if not matching_cod:
            self.log.error(f"Batch {self.batch_id} was not found in any COD* folder.")
            raise FileNotFoundError(
                f"Batch {self.batch_id} was not found in any COD* folder."
            )
        return matching_cod

    def compress_results(self, batch_data, cod):
        """Compress the analysis_results folder with a random password"""
        batch_path = batch_data["path"]
        batch = batch_data["batch"]
        analysis_dir = os.path.join(batch_path, "analysis_results")

        if not os.path.exists(analysis_dir):
            self.log.error(f"Folder analysis_results not found in {batch_path}")
            stderr.print(f"[red]Folder analysis_results not found in {batch_path}")
            return None, None, None

        passwd_7z = token_hex(8)  # Creates a random password
        filename_7z = f"{cod}_{batch}_analysis_results_{self.project}_{datetime.now().strftime('%Y%m%d%H%M%S')}.7z"
        path_7z = os.path.join(batch_path, filename_7z)

        try:
            subprocess.run(
                ["7z", "a", f"-p{passwd_7z}", "-mhe=on", path_7z, analysis_dir],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )

            self.log.info(f"Compressed file: {path_7z} with password: {passwd_7z}")
            stderr.print(
                f"[green]Compressed file: {path_7z} with password: {passwd_7z}"
            )
            return path_7z, passwd_7z, filename_7z

        except subprocess.CalledProcessError as e:
            self.log.error(f"Compression failed for {analysis_dir}: {e}")
            stderr.print(f"[red]Compression failed for {analysis_dir}")
            raise

        except Exception as e:
            self.log.error(f"Unexpected error during compression: {e}")
            stderr.print("[red]Unexpected error during compression")
            raise

    def upload_to_sftp(self, path_7z, cod):
        """Upload the compressed file to the SFTP server in the ANALYSIS_RESULTS folder inside the corresponding COD"""
        if not self.relecov_sftp.open_connection():
            self.log.error("Could not connect to SFTP server")
            stderr.print("[red]Could not connect to SFTP server")
            return False

        sftp = self.relecov_sftp.sftp

        # Define the remote path where the file is to be uploaded
        remote_dir = f"/{cod}/{self.analysis_folder}"
        remote_file_path = f"{remote_dir}/{os.path.basename(path_7z)}"

        try:
            # Check if the ANALYSIS_RESULTS folder exists inside the COD
            try:
                sftp.chdir(remote_dir)  # Try to change to the directory
            except FileNotFoundError:
                logtxt = (
                    f"Directory {remote_dir} not found. Failed to upload file to sftp."
                )
                self.log.error(logtxt)
                stderr.print(f"[red]{logtxt}")
                return False

            # Upload the compressed file
            self.log.info(f"Uploading {path_7z} to {remote_file_path} in SFTP...")
            stderr.print(f"[blue]Uploading {path_7z} to {remote_file_path} in SFTP...")
            success = self.relecov_sftp.upload_file(path_7z, remote_file_path)

            if success:
                self.log.info(f"File successfully uploaded to {remote_file_path}")
                stderr.print(f"[green]File successfully uploaded to {remote_file_path}")
                return True
            else:
                self.log.error(f"Error uploading {path_7z} file")
                stderr.print(f"[red]Error uploading {path_7z} file")
                return False

        except Exception as e:
            self.log.error(f"Error uploading file to SFTP: {e}")
            stderr.print(f"[red]Unexpected error: {e}")
            return False

        finally:
            try:
                os.remove(path_7z)
                self.log.info(f"Deleted local file {path_7z}")
            except FileNotFoundError:
                self.log.warning(f"Tried to delete {path_7z}, but it does not exist")
                stderr.print(f"[red]Tried to delete {path_7z}, but it does not exist")
            except PermissionError as e:
                self.log.error(
                    f"Permission denied when trying to delete {path_7z}: {e}"
                )
                stderr.print(
                    f"[red]Permission denied when trying to delete {path_7z}: {e}"
                )
            except Exception as e:
                self.log.error(f"Unexpected error when trying to delete {path_7z}: {e}")
                stderr.print(
                    f"[red]Unexpected error when trying to delete {path_7z}: {e}"
                )
            self.relecov_sftp.close_connection()

    def notify_lab(
        self,
        batch_name,
        lab_code,
        receiver_email,
        email_psswd=None,
        passwd_7z=None,
        filename_7z=None,
    ):
        """
        Sends an email notification to the laboratory with results details.
        """
        # Render the email content using the Jinja2 template
        email_body = self.email_sender.render_email_template(
            additional_info=f"Batch {batch_name} is ready.",
            submitting_institution_code=lab_code,
            template_name="template_results_relecov.j2",
            batch=batch_name,
            password=passwd_7z,
            zip_filename=filename_7z,
        )

        # Send the email
        try:
            self.email_sender.send_email(
                receiver_email=receiver_email,
                subject=f"Batch {batch_name} - Resultados del análisis",
                body=email_body,
                attachments=[],
                email_psswd=email_psswd,
            )
            self.log.info(
                f"Notification sent to {receiver_email} for batch {batch_name}."
            )
            stderr.print(
                f"[green]Notification sent to {receiver_email} for batch {batch_name}."
            )

        except Exception:
            self.log.error(
                f"Error while sending the email to {receiver_email} for batch {batch_name}."
            )
            stderr.print(
                f"[red]Error while sending the email to {receiver_email} for batch {batch_name}."
            )

    def execute_process(self):
        """Runs the complete flow: search, compress, upload, logging and sending the email."""
        cod_batches = self.find_cod_for_batch()

        with open(self.guide, "r") as f:
            institutions_data = json.load(f)

        for cod, batch_data in cod_batches.items():
            self.log.info(f"Processing batch {batch_data['batch']} in {cod}")
            stderr.print(f"Processing batch {batch_data['batch']} in {cod}")

            path_7z, passwd_7z, filename_7z = self.compress_results(batch_data, cod)
            if path_7z:
                success = self.upload_to_sftp(path_7z, cod)
                if success:
                    if cod not in self.processed_batches:
                        self.processed_batches[cod] = []

                    self.processed_batches[cod].append(
                        {
                            "batch": batch_data["batch"],
                            "archivo": filename_7z,
                            "contraseña": passwd_7z,
                            "fecha": datetime.now().isoformat(),
                        }
                    )
                    logtxt = f"Process completed for {batch_data['batch']} in {cod}."
                    self.log.info(logtxt)
                    stderr.print(f"[green]{logtxt}")

                    institution_info = institutions_data.get(cod)

                    if institution_info:
                        # After successful upload, send the results email
                        receiver_email = [institution_info["email_receiver"]]
                        lab_code = cod
                        self.notify_lab(
                            batch_data["batch"],
                            lab_code,
                            receiver_email,
                            passwd_7z=passwd_7z,
                            filename_7z=filename_7z,
                        )



#!/usr/bin/env python


def validate_with_exceptions(schema, data, errors):
    """Filter validation errors based on known exceptions.

    This function filters out specific validation errors from the provided list
    of errors based on exceptions defined in the schema. It allows:
      - Numeric fields (integer/float) to contain the placeholder
        'Not Provided [SNOMED:434941000124101]'.
      - String fields with format 'date' to contain the same placeholder.

    Args:
        schema (dict): Dictionary representing the JSON schema.
        data (dict): The input data being validated.
        errors (list): List of validation errors returned by a JSON schema validator.

    Returns:
        filtered_errors (list): List of validation errors excluding known exceptions.
    """
    filtered_errors = []

    for error in errors:
        property_path = ".".join(str(p) for p in error.path)
        prop_schema = schema["properties"].get(property_path, {})

        # allow not provided for numeric types
        if (
            error.validator == "type"
            and error.instance == "Not Provided [SNOMED:434941000124101]"
            and prop_schema.get("type") in ["integer", "number"]
        ):
            continue

        # allow not provided for date format types
        if (
            error.validator == "format"
            and error.instance == "Not Provided [SNOMED:434941000124101]"
            and prop_schema.get("type") == "string"
            and prop_schema.get("format") == "date"
        ):
            continue

        # Keep all other errors
        filtered_errors.append(error)

    return filtered_errors



import logging
import os
import paramiko
import rich.console
import stat
import sys
import time
from relecov_tools.config_json import ConfigJson
import relecov_tools.utils

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


class SftpRelecov:
    """Class to handle SFTP connection with remote server. It uses paramiko library to establish
    the connection. The class can be used to upload and download files from the remote server.
    The class can be initialized with a configuration file and with the username and password.
    If the configuration file is not provided, the class will try to read the configuration from
    the environment variables. If configuration file is provided, the class will read the
    configuration from the file. The format of the configuration file should be a json file with
    the following keys:
    {
        "sftp_server": "server_name",
        "sftp_port": "port_number"
    }
    """

    def __init__(self, conf_file=None, username=None, password=None):
        if conf_file is None:
            config_json = ConfigJson()
            self.sftp_server = config_json.get_topic_data("sftp_handle", "sftp_server")
            self.sftp_port = config_json.get_topic_data("sftp_handle", "sftp_port")
        else:
            config_json = conf_file
            if not os.path.isfile(conf_file):
                log.error("Configuration file %s does not exists", conf_file)
                stderr.print(
                    "[red] Configuration file does not exist. " + conf_file + "!"
                )
                sys.exit(1)
            j_data = relecov_tools.utils.read_json_file(conf_file)
            try:
                self.sftp_server = j_data["sftp_server"]
                self.sftp_port = j_data["sftp_port"]
            except KeyError as e:
                log.error("Could not find the key %s in the config file", e)
                stderr.print(
                    "[red] Could not find the key " + e + "in config file " + conf_file
                )
                sys.exit(1)
        self.user_name = username
        self.password = password
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    def reconnect_if_fail(n_times, sleep_time):
        def decorator(func):
            def retrier(self, *args, **kwargs):
                more_sleep_time = 0
                retries = 0
                while retries < n_times:
                    try:
                        return func(self, *args, **kwargs)
                    except Exception:
                        retries += 1
                        log.info("Connection lost. Trying to reconnect...")
                        time.sleep(more_sleep_time)
                        # Try extending sleep time before reconnecting in each step
                        more_sleep_time = more_sleep_time + sleep_time
                        self.open_connection()
                else:
                    log.error("Could not reconnect to remote client")
                return func(self, *args, **kwargs)

            return retrier

        return decorator

    def open_connection(self):
        """Establishing sftp connection"""
        log.info("Setting credentials for SFTP connection with remote server")
        self.client.connect(
            hostname=self.sftp_server,
            port=self.sftp_port,
            username=self.user_name,
            password=self.password,
            allow_agent=False,
            look_for_keys=False,
        )
        try:
            log.info("Trying to establish SFTP connection")
            self.sftp = self.client.open_sftp()
        except Exception as e:
            log.error("Could not establish SFTP connection: %s", e)
            stderr.print("[red]Could not establish SFTP connection")
            return False
        return True

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def list_remote_folders(self, folder_name, recursive=False):
        """Creates a directories list from the given client remote path

        Args:
            folder_name (str): folder name in remote path
            recursive (bool, optional): finds all subdirectories too. Defaults to False.

        Returns:
            directory_list(list(str)): Names of all folders within remote folder
        """
        log.info("Listing directories in %s", folder_name)
        directory_list = []
        try:
            content_list = self.sftp.listdir_attr(folder_name)
            subfolders = any(stat.S_ISDIR(item.st_mode) for item in content_list)
        except (FileNotFoundError, OSError) as e:
            log.error("Invalid folder at remote sftp %s", e)
            raise
        if not subfolders:
            return [folder_name]

        def recursive_list(folder_name):
            try:
                attribute_list = self.sftp.listdir_attr(folder_name)
            except (FileNotFoundError, OSError) as e:
                log.error("Invalid folder at remote sftp %s", e)
                raise
            for attribute in attribute_list:
                if stat.S_ISDIR(attribute.st_mode):
                    abspath = os.path.join(folder_name, attribute.filename)
                    directory_list.append(abspath)
                    recursive_list(abspath)
                else:
                    continue
            return directory_list

        if recursive:
            log.debug("Listing recursive")
            directory_list = recursive_list(folder_name)
            if folder_name != ".":
                directory_list.append(folder_name)
            return directory_list
        try:
            log.debug("Listing content in folders...")
            directory_list = [
                item.filename for item in content_list if stat.S_ISDIR(item.st_mode)
            ]
        except AttributeError:
            return False
        self.close_connection()
        return directory_list

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def get_file_list(self, folder_name, recursive=False):
        """Return a tuple with file name and directory path from remote

        Args:
            folder_name (str): name of folder in remote repository
            recursive (bool): either to list files recursively through subfolders

        Returns:
            file_list (list(str)): list of files in remote folder
        """
        log.info("Listing files in %s", folder_name)
        file_list = []
        try:
            content_list = self.sftp.listdir_attr(folder_name)
            for content in content_list:
                full_path = os.path.join(folder_name, content.filename)
                if stat.S_ISDIR(content.st_mode):
                    if recursive is True:
                        file_list.extend(self.get_file_list(full_path))
                elif stat.S_ISREG(content.st_mode):
                    file_list.append(full_path)
        except FileNotFoundError as e:
            log.error(f"Folder not found: {folder_name}")
            raise e
        return file_list

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def get_from_sftp(self, file, destination, exist_ok=False):
        """Download a file from remote sftp

        Args:
            file (str): path of the file in remote sftp
            destination (str): local path of the file after download
            exist_ok (bool): Skip download if file exists in local destination

        Returns:
            bool: True if download was successful, False if it was not
        """
        if os.path.exists(destination) and exist_ok:
            return True
        else:
            try:
                self.sftp.get(file, destination)
                return True
            except FileNotFoundError as e:
                log.error("Unable to fetch file %s ", e)
                return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def make_dir(self, folder_name):
        """Create a new directory in remote sftp

        Args:
            folder_name (str): name of the directory to be created

        Returns:
            bool: True if directory was created, False if it was not
        """
        try:
            self.sftp.mkdir(folder_name)
            return True
        except FileExistsError:
            log.error("Directory %s already exists", folder_name)
            stderr.print("[red]Directory already exists")
            return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def rename_file(self, old_name, new_name):
        """Rename a file in remote sftp

        Args:
            old_name (str): current name of the file
            new_name (str): new name of the file

        Returns:
            bool: True if file was renamed, False if it was not
        """
        try:
            self.sftp.rename(old_name, new_name)
            return True
        except FileNotFoundError as e:
            error_txt = f"Could not rename {old_name} to {new_name}: {e}"
            log.error(error_txt)
            stderr.print(f"[red]{error_txt}")
            return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def remove_file(self, file_name):
        """Remove a file from remote sftp

        Args:
            file_name (str): name of the file to be removed

        Returns:
            bool: True if file was removed, False if it was not
        """
        try:
            self.sftp.remove(file_name)
            log.info("%s Deleted from remote server", file_name)
            return True
        except FileNotFoundError:
            log.error("File %s not found", file_name)
            stderr.print("[red]File not found")
            return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def remove_dir(self, folder_name):
        """Remove a directory from remote sftp

        Args:
            folder_name (str): name of the directory to be removed

        Returns:
            bool: True if directory was removed, False if it was not
        """
        try:
            self.sftp.rmdir(folder_name)
            return True
        except FileNotFoundError:
            log.error("Directory %s not found", folder_name)
            stderr.print("[red]Directory not found")
            return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def upload_file(self, local_path, remote_file):
        """Upload a file to remote sftp

        Args:
            localpath (str): path of the file in local machine
            remote_file (str): path of the file in remote sftp

        Returns:
            bool: True if file was uploaded, False if it was not
        """
        try:
            self.sftp.put(local_path, remote_file)
            return True
        except FileNotFoundError as e:
            log.error("File not found %s", e)
            stderr.print("[red]File not found")
            return False

    @reconnect_if_fail(n_times=3, sleep_time=30)
    def close_connection(self):
        log.info("Closing SFTP connection")
        try:
            self.sftp.close()
        except NameError:
            log.warning("Could not close sftp connection")
            return False
        log.info("SFTP connection closed")
        return True



#!/usr/bin/env python
"""
Common utility function used for relecov_tools package.
"""
import os
import sys
import glob
import hashlib
import logging
import questionary
import json
import openpyxl
import yaml
import gzip
import re
import shutil
from itertools import islice, product
from Bio import SeqIO
from rich.console import Console
from rich.table import Table
from datetime import datetime
from tabulate import tabulate
from secrets import token_hex
import openpyxl.utils
import openpyxl.styles
import pandas as pd
import semantic_version
import subprocess
import importlib.metadata

log = logging.getLogger(__name__)


def file_exists(file_to_check):
    """
    Input:
        file_to_check   # file name to check if exists
    Return:
        True if exists
    """
    if os.path.isfile(file_to_check):
        return True
    return False


def safe_remove(file_path):
    try:
        os.remove(file_path)
    except OSError:
        return False
    return True


def get_files_match_condition(condition):
    """find all path names that matches with the condition"""
    return glob.glob(condition)


def read_json_file(j_file):
    """Read json file."""
    with open(j_file, "r") as fh:
        try:
            data = json.load(fh)
        except (UnicodeDecodeError, ValueError):
            raise

    return data


def read_excel_file(f_name, sheet_name, header_flag, leave_empty=True):
    """Read the input excel file and return the data as a list of dictionaries.
    If openpyxl fails, fall back to pandas but return in the same format.
    """
    try:
        wb_file = openpyxl.load_workbook(f_name, data_only=True)
        ws_metadata_lab = wb_file[sheet_name]
        try:
            heading_row = [
                idx + 1
                for idx, x in enumerate(ws_metadata_lab.values)
                if header_flag in x
            ][0]
        except IndexError:
            raise KeyError(
                f"Header flag '{header_flag}' could not be found in {f_name}"
            )
        heading = [
            str(i.value).strip() for i in ws_metadata_lab[heading_row] if i.value
        ]
        ws_data = []
        for row in islice(ws_metadata_lab.values, heading_row, ws_metadata_lab.max_row):
            l_row = list(row)
            if all(cell is None for cell in l_row):
                continue
            data_row = {}
            for idx in range(0, len(heading)):
                if l_row[idx] is None:
                    data_row[heading[idx]] = (
                        None if leave_empty else "Not Provided [SNOMED:434941000124101]"
                    )
                else:
                    data_row[heading[idx]] = l_row[idx]
            ws_data.append(data_row)
        return ws_data, heading_row

    except Exception as e:
        try:
            df = pd.read_excel(f_name, sheet_name=sheet_name, header=None)

            heading_row_idx = df.apply(lambda row: header_flag in row.values, axis=1)
            if not heading_row_idx.any():
                raise KeyError(
                    f"Header flag '{header_flag}' could not be found in {f_name}"
                )
            heading_row = heading_row_idx.idxmax()
            heading = [str(h).strip() for h in df.iloc[heading_row] if pd.notna(h)]

            ws_data = []
            for _, row in df.iloc[heading_row + 1 :].iterrows():
                if row.isna().all():
                    continue
                data_row = {}
                for idx in range(len(heading)):
                    val = row.iloc[idx] if idx < len(row) else None
                    if pd.isna(val):
                        data_row[heading[idx]] = (
                            None
                            if leave_empty
                            else "Not Provided [SNOMED:434941000124101]"
                        )
                    else:
                        data_row[heading[idx]] = val
                ws_data.append(data_row)
            return (
                ws_data,
                heading_row + 1,
            )  # +1 to maintain consistency with openpyxl (1-based)

        except Exception as fallback_e:
            raise RuntimeError(
                f"Failed to read file with both openpyxl and pandas:\n- openpyxl error: {e}\n- pandas error: {fallback_e}"
            )


def string_to_date(string):
    """Convert date (Y-M-D...) from string to date. Tries iteratively with variable
    number of digits and multiple separators, starting from seconds up to year.
    args:
        string (str): Date in string format to be parsed: e.g. 2020-08-07-12-00-00
    returns:
        res_date (datetime.datetime): String converted to date format
    """

    def rec_date_extraction(string, digits, sep):
        regex = r"^\d{4}"  # The string should start from year
        for _ in range(0, digits - 4, 2):
            new_reg = sep + "\d{2}"  # Each date param occupies 2 digits (4 for year)
            regex = regex + new_reg
        match = re.match(regex, string)
        if not match:
            match = re.match(regex, string.replace(":", sep))
            if not match:
                raise ValueError(f"Could not match date to given string: {string}")
        matchdate = match.group(0)
        # Getting the equivalent format from given digits
        full_date = "%Y%m%d%H%M%S"[0 : digits - 2]
        datepattern = f"{sep}%".join(full_date.split("%")).strip(sep)
        res_date = datetime.strptime(matchdate, datepattern)
        return res_date

    seps = ["", " ", "/", "-", "_"]
    digits_list = [x for x in range(4, 16, 2)]
    combinations = sorted(product(digits_list, seps), reverse=True)
    for digits, sep in combinations:
        try:
            res_date = rec_date_extraction(string, digits, sep)
        except ValueError:
            continue
        return res_date
    else:
        return None


def excel_date_to_num(date):
    """Transform a date object formatted by excel to a numeric value"""
    try:
        return date.toordinal() - datetime(1899, 12, 30).toordinal()
    except AttributeError:
        return None


def read_csv_file_return_dict(file_name, sep=None, key_position=0):
    """Read csv or tsv file, according to separator (sep), and return a dictionary
    where the main key is the first column, if key position is None otherwise
    the index value of the key position is used as key. If sep is None then
    try to assert a separator automatically depending on file extension.
    """
    if sep is None:
        file_extension = os.path.splitext(file_name)[1]
        extdict = {".csv": ",", ".tsv": "\t", ".tab": "\t"}
        # Use space as a default separator, None would also be valid
        sep = extdict.get(file_extension, " ")
    try:
        # Read all columns as strings to avoid parsing IDs as float buy try to infer datatypes afterwards
        file_df = pd.read_csv(file_name, sep=sep, dtype="string").convert_dtypes()
    except FileNotFoundError:
        raise
    key_column = file_df.columns[key_position]
    file_data = file_df.set_index(key_column).to_dict(orient="index")
    return file_data


def read_fasta_return_SeqIO_instance(file_name):
    """Read fasta and return SeqIO instance"""
    try:
        return SeqIO.read(file_name, "fasta")
    except FileNotFoundError:
        raise


def read_yml_file(file_name):
    """Read yml file"""
    file_name = os.path.expanduser(file_name)
    with open(file_name, "r") as fh:
        try:
            return yaml.safe_load(fh)
        except yaml.YAMLError:
            raise


def get_md5_from_local_folder(local_folder):
    """Fetch the md5 values for each file in the file list"""
    md5_results = {}
    reg_for_md5 = os.path.join(local_folder, "*.md5")
    # reg_for_non_md5 = os.path.join(local_folder, "*[!.md5]")
    md5_files = glob.glob(reg_for_md5)
    if not md5_files:
        return False
    else:
        for md5_file in md5_files:
            file_path_name, f_ext = os.path.splitext(md5_file)
            if not file_exists(file_path_name):
                log.error("Found md5 file but not %s", file_path_name)
                continue
            file_name = os.path.basename(file_path_name)
            fh = open(md5_file, "r")
            md5_results[file_name] = fh.read()
            fh.close()
    return md5_results


def read_md5_checksum(file_name, avoid_chars=list()):
    """Read MD5_checksum file and return a dict of {file: md5_hash}

    Args:
        file_name (str): file containing "md5hash  file" in tab separated format
        avoid_chars (list(str), optional): Lines with any of these elements
        will be skipped. Defaults to list().

    Returns:
        hash_dict(dict): dictionary of {file: md5_hash}
    """
    try:
        with open(file_name, "r") as file:
            content = file.read()
    except FileNotFoundError:
        raise
    clean_content = content.replace("*", "")
    lines = clean_content.splitlines()
    translation = str.maketrans("", "", "'\"")
    if any("\t" in line for line in lines):
        lines = [line.strip().translate(translation).split("\t") for line in lines]
    elif any("," in line for line in lines):
        lines = [line.strip().translate(translation).split(",") for line in lines]
    else:
        lines = [line.strip().translate(translation).split() for line in lines]
    clean_lines = [
        x for x in lines if not any(ch in string for ch in avoid_chars for string in x)
    ]
    # md5sum should always have 2 columns: hash - path
    md5_lines = [line for line in clean_lines if len(line) == 2]
    if not md5_lines:
        return False
    # split paths for both windows "\" and linux "/" using regex [\\/]
    hash_dict = {re.split(r"[\\/]", line[1])[-1]: line[0].lower() for line in md5_lines}
    return hash_dict


def delete_local_folder(folder):
    """Delete download folder because files does not complain requisites"""
    log.info("Deleting local folder %s", folder)
    shutil.rmtree(folder, ignore_errors=True)
    return True


def calculate_md5(file_name):
    """Calculate the md5 value for the file name"""
    return hashlib.md5(open(file_name, "rb").read()).hexdigest()


def write_md5_file(file_name, md5_value):
    """Write md5 to file"""
    with open(file_name, "w") as fh:
        fh.write(md5_value)
    return


def create_md5_files(local_folder, file_list):
    """Create the md5 files and return their value"""
    md5_results = {}
    for file_name in file_list:
        md5_results[file_name] = [
            local_folder,
            calculate_md5(os.path.join(local_folder, file_name)),
        ]
        md5_file_name = file_name + ".md5"
        write_md5_file(
            os.path.join(local_folder, md5_file_name), md5_results[file_name][1]
        )
    return md5_results


def save_local_md5(file_name, md5_value):
    """Save the MD5 value"""
    with open(file_name, "w") as fh:
        fh.write(md5_value)
    return True


def write_json_to_file(data, file_name):
    """Write metadata to json file"""
    with open(file_name, "w", encoding="utf-8") as fh:
        fh.write(json.dumps(data, indent=4, sort_keys=True, ensure_ascii=False))
    return True


def compress_file(file):
    """compress a given file with gzip, adding .gz extension afterwards

    Args:
        file (str): path to the given file
    """
    try:
        with open(file, "rb") as raw, gzip.open(f"{file}.gz", "wb") as comp:
            comp.writelines(raw)
        return True
    except FileNotFoundError:
        return False


def check_gzip_integrity(file_path):
    """Check if a compressed file can be decompressed"""
    chunksize = 100000000  # 10 Mbytes
    with gzip.open(file_path, "rb") as f:
        try:
            while f.read(chunksize) != b"":
                pass
        except Exception:
            # Not a gzip file
            return False
        # EOFError: Compressed file is truncated
        except EOFError:
            return False
    return True


def lower_keys(data):
    """Transform all keys to lowercase strings in a dictionary"""
    return {str(key).lower(): v for key, v in data.items()}


def rich_force_colors():
    """
    Check if any environment variables are set to force Rich to use coloured output
    """
    if (
        os.getenv("GITHUB_ACTIONS")
        or os.getenv("FORCE_COLOR")
        or os.getenv("PY_COLORS")
    ):
        return True
    return None


stderr = Console(
    stderr=True, style="dim", highlight=False, force_terminal=rich_force_colors()
)


def prompt_text(msg):
    source = questionary.text(msg).unsafe_ask()
    return source


def prompt_password(msg):
    source = questionary.password(msg).unsafe_ask()
    return source


def prompt_tmp_dir_path():
    stderr.print("Temporal directory destination to execute service")
    source = questionary.path("Source path").unsafe_ask()
    return source


def prompt_selection(msg, choices):
    selection = questionary.select(msg, choices=choices).unsafe_ask()
    return selection


def prompt_path(msg):
    source = questionary.path(msg).unsafe_ask()
    return source


def prompt_yn_question(msg):
    confirmation = questionary.confirm(msg).unsafe_ask()
    return confirmation


def prompt_skip_folder_creation():
    stderr.print("Do you want to skip folder creation? (Y/N)")
    confirmation = questionary.confirm("Skip?", default=False).unsafe_ask()
    return confirmation


def prompt_checkbox(msg, choices):
    selected_options = questionary.checkbox(msg, choices=choices).unsafe_ask()
    return selected_options


def get_file_date(file_path):
    """Get the modification date of a file."""
    try:
        # Get the modification time of the file
        mtime = os.path.getmtime(file_path)
        # Convert the modification time to a datetime object
        file_date = datetime.fromtimestamp(mtime)
        # Format date
        formatted_date = file_date.strftime("%Y%m%d")
        return formatted_date
    except FileNotFoundError:
        # Handle file not found error
        print(f"File not found: {file_path}")
        return None


def select_most_recent_files_per_sample(paths_list):
    """Selects the most recent file for each sample among potentially duplicated files.
    Input:
        - paths_list: a list of sample's file paths.
    Output:
        - List of file paths containig the most recent/up-to-date file for each sample.
    """
    filename_groups = {}
    # Count occurrences of each filename and group files by sample names
    for file in paths_list:
        # TODO: So far, it uses split method to identify this pattern:
        # [sample1.pangolin.csv, sample1.pangolin_20240310.csv]. It should be
        # improve to parse files based on a different character matching field.
        file_name = os.path.basename(file).split(".")[0]
        if file_name in filename_groups:
            filename_groups[file_name].append(file)
        else:
            filename_groups[file_name] = [file]
    # Filter out sample names with only one file
    duplicated_files = [
        (sample_name, file_paths)
        for sample_name, file_paths in filename_groups.items()
        if len(file_paths) > 1
    ]
    # Iterate over duplicated files to select the most recent one for each sample
    for sample_name, file_paths in duplicated_files:
        stderr.print(
            f"\tMore than one file found for sample {sample_name}. Selecting the most recent one."
        )
        # Sort files by modification time (most recent first)
        sorted_files = sorted(
            file_paths, key=lambda file_path: os.path.getmtime(file_path), reverse=True
        )
        # Select the most recent file
        selected_file = sorted_files[0]
        stderr.print(f"\tSelected file for sample {sample_name}: {selected_file}")
        # Remove other files for the same sample from the filtered_files dictionary
        filename_groups[sample_name] = [selected_file]
    # Update filename_groups with filtered files
    filename_groups = [
        (sample_name, file_path)
        for sample_name, file_paths in filename_groups.items()
        for file_path in file_paths
    ]
    # Reformat variable to retrieve a list of file paths
    file_path_list = [sample_file_path for _, sample_file_path in filename_groups]
    return file_path_list


def print_log_report(
    log_report, categories=None, sections=["warning", "valid", "error"]
):
    color_codes = {
        "error": "\033[91m",  # Red
        "warning": "\033[93m",  # Orange
        "valid": "\033[92m",  # Green
        "reset": "\033[0m",  # Reset color
    }
    table_data = []
    for section_name, section_data in log_report.items():
        if section_name in sections:
            for category, items in section_data.items():
                if categories is None or category in categories:
                    colored_category = (
                        f"{color_codes[section_name]}{category}{color_codes['reset']}"
                    )
                    for item in items:
                        colored_message = (
                            f"{color_codes[section_name]}{item}{color_codes['reset']}"
                        )
                        table_data.append(
                            [section_name, colored_category, colored_message]
                        )
    print(
        tabulate(
            table_data,
            headers=["Log type", "Category", "Message"],
            tablefmt="presto",
        )
    )


def prompt_create_outdir(
    path, folder_name=None, prompt_message="Define path to store the output:"
):
    """Ensure the directory exists or prompt the user to define and create it."""
    # Check path
    if not path:
        path = prompt_path(prompt_message)
        stderr.print(f"Chosen directory: {path}")

    # Check folder_name
    if not folder_name:
        default_folder = prompt_yn_question(
            "Do you want to use the default directory ('results') to store the results? (yes/no):"
        )
        if not default_folder:
            folder_name = prompt_text("Write your output directory: ")
        else:
            folder_name = "results"

    # Prevent duplicate folder names in the path
    if os.path.basename(path) == folder_name:
        global_path = path
    else:
        global_path = os.path.join(path, folder_name)

    if not os.path.exists(global_path):
        create_folder = prompt_yn_question(
            f"The directory does not exist. Do you want to create '{folder_name}' folder in this path? (yes/no):"
        )
        if create_folder:
            os.makedirs(global_path)
            stderr.print(f"[green]Folder '{folder_name}' created at {path}")
        else:
            stderr.print("[red]Directory creation aborted.")
            sys.exit(1)
    elif os.path.isdir(global_path):
        os.makedirs(global_path, exist_ok=True)
        stderr.print(f"[green]Defining '{folder_name}' as output folder")
    else:
        stderr.print("[red]The provided path is not a directory.")
        sys.exit(1)

    return global_path


def adjust_sheet_size(sheet, wrap_text=True, col_width=30):
    """Adjust column width and row heights depending on the max number of
    characters in each one.

    Args:
        sheet (openpyxl.worksheet): active openpyxl worksheet object
        wrap_text (bool): Wether to use excel wrap_text function for each cell. Defaults to True
        col_width (int): Minimum columns width value. Also used to define maximum
        number of characters in each cell when wrap_text is True. Defaults to 30.
    """
    dims = {}
    for _, row in enumerate(sheet.iter_rows(min_row=2, max_row=sheet.max_row), start=2):
        for cell in row:
            if wrap_text:
                cell.alignment = openpyxl.styles.Alignment(wrapText=True)
            if cell.value is not None:
                max_length = max((dims.get(cell.column, 0), len(str(cell.value))))
                dims[cell.column] = max_length / col_width
    for col_num, value in dims.items():
        if value < col_width:
            value = col_width
        sheet.column_dimensions[openpyxl.utils.get_column_letter(col_num)].width = value


def validate_semantic_version(version):
    try:
        ver = semantic_version.Version(version)
        return ver
    except ValueError:
        return None


def get_package_name():
    """Get project name"""
    try:
        package_name = importlib.metadata.metadata(__name__.split(".")[0])["Name"]
        return package_name
    except importlib.metadata.PackageNotFoundError:
        return "unknown_package"


def get_git_branch():
    """Get current git branch"""
    try:
        branch = (
            subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
            .strip()
            .decode("utf-8")
        )
        return branch
    except Exception:
        return "main"  # if not able to retrieve git branch, add fixed value


def get_schema_url():
    """Generates the schema url dinamically"""
    package_name = get_package_name()
    branch_name = get_git_branch()
    base_url = "https://github.com/BU-ISCIII"
    schema_path = f"{package_name}/blob/{branch_name}/{package_name.replace('-','_')}/schema/relecov_schema.json"

    return f"{base_url}/{schema_path}"


def display_dataframe_to_user(name: str, dataframe: pd.DataFrame):
    """
    Display a Pandas DataFrame in a formatted table using Rich.

    Args:
        name (str): Title of the table.
        dataframe (pd.DataFrame): The DataFrame to display.
    """
    console = Console()

    # If DataFrame is empty, show a message
    if dataframe.empty:
        console.print(f"[bold red]{name} - No Data Available[/bold red]")
        return

    # Create a Rich Table
    table = Table(title=name, show_lines=True)

    # Add columns
    for col in dataframe.columns:
        table.add_column(col, justify="left", style="cyan", no_wrap=True)

    # Add rows
    for _, row in dataframe.iterrows():
        table.add_row(*[str(value) for value in row])

    # Display the table
    console.print(table)


def load_schema(schema_path: str) -> dict:
    """
    Load a JSON schema from the specified file path.

    Args:
        schema_path (str): The file path to the JSON schema.

    Returns:
        dict: Parsed schema as a Python dictionary.
    """
    with open(schema_path) as f:
        return json.load(f)


def get_available_software(json_path: str) -> list:
    """
    Retrieve available software names from a bioinfo configuration JSON file.

    Args:
        json_path (str): Path to the bioinfo configuration file.

    Returns:
        list: A list of available software/tools defined in the configuration.
    """
    config = read_json_file(json_path)
    return list(config.keys())


def cast_value_to_schema_type(value, expected_type: str):
    """
    Cast a value to the expected JSON schema type.

    Args:
        value (any): The input value to be cast.
        expected_type (str): Target data type from the schema. Options: "integer", "number", "boolean", "string".

    Returns:
        any: The value cast to the appropriate type, or a string fallback if casting fails.
    """
    if expected_type == "integer":
        try:
            return int(float(value))
        except (ValueError, TypeError):
            return str(value).strip()
    elif expected_type == "number":
        try:
            return float(value)
        except (ValueError, TypeError):
            return str(value).strip()
    elif expected_type == "boolean":
        return str(value).strip().lower() in ["true", "yes", "1"]
    elif expected_type == "string":
        return str(value).strip()
    else:
        try:
            return str(value).strip()
        except Exception:
            return value


def get_safe_hex(output_dir, length=3):
    """Return an unique hexadecimal code that does not repeat in any file
    of the given output directory

    Args:
        output_dir (str): Folder where files will be checked
        length (int): Number of bytes for the code. Character length will be doubled

    Returns:
        hex_id (str): Non-repeating hexadecimal code
    """

    def get_new_hex(hex_id, output_dir):
        """Recursive search for new hexadecimal codes if exist"""
        if any(hex_id in x for x in os.listdir(output_dir)):
            return get_new_hex(token_hex(length).upper(), output_dir)
        else:
            return token_hex(length).upper()

    hex_id = get_new_hex(token_hex(length).upper(), output_dir)
    return hex_id



#!/usr/bin/env python
import os
import logging
import rich
import os.path
import pandas as pd

import relecov_tools.utils
import relecov_tools.assets.pipeline_utils.utils
from relecov_tools.read_bioinfo_metadata import BioinfoReportLog

log = logging.getLogger(__name__)
stderr = rich.console.Console(
    stderr=True,
    style="dim",
    highlight=False,
    force_terminal=relecov_tools.utils.rich_force_colors(),
)


def quality_control_evaluation(data):
    """Evaluates QC status for each sample based on predefined thresholds.

    Parameters:
    -----------
    data : list of dict
        List of sample metadata dictionaries containing metrics such as coverage,
        ambiguity, number of Ns, %LDMutations, etc.

    Returns:
    --------
    list of dict
        The same list with an added 'qc_test' field per sample:
        - 'pass' if all evaluable conditions are met
        - 'fail' if any condition fails
        - Adds 'qc_failed' field with reasons for failure
    """
    log_report = BioinfoReportLog()
    method_name = f"{quality_control_evaluation.__name__}"

    thresholds = {
        "number_of_unambiguous_bases": (">", 11000),
        "number_of_Ns": ("<", 2300),
        "pass_reads": (">", 23000),
        "per_reads_host": ("<", 20.0),
        "per_hagene_coverage": (">", 98.0),
        "per_nagene_coverage": (">", 98.0),
    }

    def is_not_evaluable(value):
        return False

    def invert_operator(op):
        return {">": "<", "<": ">", ">=": "<=", "<=": ">=", "==": "!=", "!=": "=="}.get(
            op, f"NOT_{op}"
        )

    conditions = {
        k: (
            lambda x, op=op, th=th: (
                eval(f"{float(x)} {op} {th}") if isinstance(x, (int, float)) else False
            )
        )
        for k, (op, th) in thresholds.items()
    }

    data, warnings = relecov_tools.assets.pipeline_utils.utils.evaluate_qc_samples(
        data,
        thresholds,
        conditions,
        invert_operator,
        is_not_evaluable,
    )

    for warn in warnings:
        log_report.update_log_report(method_name, warn)

    return data


def get_software_versions(files_list, file_tag, pipeline_name, output_dir=None):
    """File handler to parse software versions from csv.

    Args:
        files_list (list): A list with paths to software version files.

    Returns:
        software_versions: A dictionary containing software versions.
    """
    method_name = f"{get_software_versions.__name__}"
    method_log_report = BioinfoReportLog()

    version_dict = {}
    for file in files_list:
        if os.path.isfile(file):
            df = pd.read_csv(file)
            for _, row in df.iterrows():
                name = row["software_name"]
                version = row["software_version"]
                id = row["software_name"].strip().lower()
                if "/" not in id:
                    id = id.split()[0]
                version_dict[id] = {"software_version": version, "software_name": name}
        else:
            method_log_report.update_log_report(
                method_name,
                "warning",
                f"Verions file {file} not found. Skipping.",
            )
    return version_dict
